import numpy as np
import numpy.ma as mp
import logging
import sys

from xframe.library import pythonLibrary as pyLib
from xframe.library.gridLibrary import NestedArray
from xframe.library.mathLibrary import spherical_to_cartesian
from xframe.library.mathLibrary import cartesian_to_spherical
from xframe.library.mathLibrary import _pi_in_q__to__reciprocity_coefficient
from xframe.library.pythonLibrary import get_L2_cache_split_parameters
from xframe.library import mathLibrary as mLib
from xframe.interfaces import DatabaseInterface,PresenterInterface
from xframe.presenters.matplotlibPresenter import heatPolar2D
from xframe import settings
pres=heatPolar2D()
log=logging.getLogger('root')
array=np.array

def getAnalysisRecipeFacotry(database,presenters):
    def getAnalysisOperatorDict(database,presenters):
        def getDataBaseOperatorDict(database):
            interfaceFunctionNames=pyLib.getFunctionNames(DatabaseInterface)
            allMethodsDict=pyLib.getMethods(database, _type='dict')
            dataDict={functionName+'@Database':allMethodsDict[functionName] for functionName in interfaceFunctionNames}
            return dataDict
        
        def getPresenterOperatorDict(presenters):
            interfaceFunctionNames=pyLib.getFunctionNames(PresenterInterface)
            presenterDict={}
            for presenter in presenters:
                Name=type(presenter).__name__
                methodsDict=pyLib.getMethods(presenter, _type='dict')
                tempDict={functionName+'@'+Name:methodsDict[functionName] for functionName in interfaceFunctionNames}
                presenterDict.update(tempDict)
            return presenterDict
    
        operatorDict={}
        analysisLibModule=__import__('analysisLibrary')  #fromlist is neccessary otherwise it only imports toplevel module analysis instead of analysis.analysusLibrary
        libraryDict=pyLib.getFunctions(analysisLibModule, _type='dict',recoursive=True)
#        print('this is the libraryDict: {} \n'.format(libraryDict))
#        print('current Module name = {}'.format(currentModule))
        operatorDict.update(libraryDict)
        mathDict=pyLib.getFunctions(mLib, _type='dict')
        operatorDict.update(mathDict)
        dataDict=getDataBaseOperatorDict(database)
        operatorDict.update(dataDict)        
        presenterDict=getPresenterOperatorDict(presenters)
        operatorDict.update(presenterDict)        
        return operatorDict

    operatorDict= getAnalysisOperatorDict(database,presenters)
    recipeFactory=pyLib.RecipeFactory(operatorDict)
    return recipeFactory


def get_analysis_process_factory(database,presenters):
    def getAnalysisOperatorDict(database,presenters):
        def getDataBaseOperatorDict(database):
            interfaceFunctionNames=pyLib.getFunctionNames(DatabaseInterface)
            allMethodsDict=pyLib.getMethods(database, _type='dict')
            dataDict={functionName+'@Database':allMethodsDict[functionName] for functionName in interfaceFunctionNames}
            return dataDict
        
        def getPresenterOperatorDict(presenters):
            interfaceFunctionNames=pyLib.getFunctionNames(PresenterInterface)
            presenterDict={}
            for presenter in presenters:
                Name=type(presenter).__name__
                methodsDict=pyLib.getMethods(presenter, _type='dict')
                tempDict={functionName+'@'+Name:methodsDict[functionName] for functionName in interfaceFunctionNames}
                presenterDict.update(tempDict)
            return presenterDict
    
        operatorDict={}
        #from .. import analysisLibrary as analysisLibModule
        
        libraryDict=pyLib.getFunctions(sys.modules[__name__], _type='dict',recoursive=True)
#        print('this is the libraryDict: {} \n'.format(libraryDict))
#        print('current Module name = {}'.format(currentModule))
        operatorDict.update(libraryDict)
        mathDict=pyLib.getFunctions(mLib, _type='dict')
        operatorDict.update(mathDict)
        #dataDict=getDataBaseOperatorDict(database)
        #operatorDict.update(dataDict)        
        #presenterDict=getPresenterOperatorDict(presenters)
        #operatorDict.update(presenterDict)        
        return operatorDict

    operatorDict= getAnalysisOperatorDict(database,presenters)
    recipeFactory=pyLib.RecipeFactory(operatorDict)
    return recipeFactory



def returnGrid_deleteFirstDimension(oldGrid):
    def deleteFirstCoordinate(point):
        return np.delete(point,0)
    reducedGridArray=np.array(tuple(map(deleteFirstCoordinate,oldGrid[0]))).reshape(oldGrid.array.shape[1:]+(oldGrid.dimension-1,))
    reducedGridArray=pyLib.getArrayOfArray(reducedGridArray)
    reducedGrid=Grid(reducedGridArray,oldGrid.gridType)
    return reducedGrid

def save_to_dict(dictionary,keys,mode,data):
    if not isinstance(keys,list):
        keys = [keys]
    folder=dictionary
    try:
        for key in keys[:-1]:
            folder = folder[key]
        key = keys[-1]
        if mode == 'append':
            dataset = folder.get(key,[])
            dataset.append(data)
            folder[key]=dataset
        elif mode=='overwrite':            
            folder[key] = data
        elif mode == 'iterative_append':
            folder = folder[key]
            #log.info('saving {}'.format(data))
            #log.info('to dict {}'.format(folder))
            for dkey in data:                
                dataset = folder.get(dkey,[])
                dataset.append(data[dkey])
                folder[dkey]=dataset
        elif mode == 'iterative_overwrite':
            folder = folder[key]
            for dkey in data:                
                folder[dkey]=data[dkey]
        else:
            folder[key] = data
    except AttributeError as e:
        log.error('Saving to dictionary {} failed with keys {}.'.format(dictionary,keys))            
    return data

def load_from_dict(dictionary,keys):
    if not isinstance(keys,list):
        keys = [keys]
    value=dictionary
    try:
        for key in keys:
            value = value[key]
    except AttributeError as e:
        log.error('loading from dictionary {} failed with keys {}.'.format(dictionary,keys))            
    return value

def generate_load_from_dict(dictionary):
    def load_from_dict(key):
        value=dictionary[key]
        return value
    return load_from_dict

        
def copy(data):
    #np.array creates copy by default and is slightly faster than data.copy()
    return array(data)


def _generate_square_default(out_array):
    mult = np.multiply
    np_square = np.square
    if out_array.dtype == np.dtype(complex):
        def square(data):
            return mult(data,data.conj(),out = out_array)
    else:
        def square(data):
            return np_square(data,out = out_array)
    return square
def _generate_square_cache_aware(out_array,L2_cache):
    data_shape = out_array.shape
    data_type = out_array.dtype
    splitting_dimension,step = get_L2_cache_split_parameters(data_shape,data_type,L2_cache)

    mult = np.multiply
    
    def square_1_loop(data):
        for i in range(0,data_shape[0],step):
            i2 = i+step
            d = data[i:i2]
            o = out_array[i:i2]
            mult(d,d.conj(),out = o)
        return out_array
    def square_2_loop(data):
        for i in range(data_shape[0]):
            for j in range(0,data_shape[1],step):
                j2 = j+step
                d = data[i,j:j2]
                o = out_array[i,j:j2]
                mult(d,d.conj(),out = o)
        return out_array
    def square_3_loop(data):
        for i in range(data_shape[0]):
            for j in range(data_shape[1]):
                for k in range(0,data_shape[2],step):
                    k2 = k+step
                    d = data[i,j,k:k2]
                    o = out_array[i,j,k:k2]
                    mult(d,d.conj(),out = o)
        return out_array
    if (splitting_dimension <= -1) or (out_array.dtype != np.dtype(complex)):
        square = _generate_square_default(out_array)
    elif splitting_dimension == 0:
        square = square_1_loop
    elif splitting_dimension == 1:
        square = square_2_loop
    elif splitting_dimension == 2:
        square = square_3_loop
    return square


def generate_square(data_shape,data_type,cache_aware=False,L2_cache=256):
    out = np.zeros(data_shape,dtype = data_type)
    if not cache_aware:
        square = _generate_square_default(out)
    else:
        square = _generate_square_cache_aware(out,L2_cache)
    return square



def _generate_abs_value_default(out_array):
    sqrt = np.sqrt
    def abs_value(data):
        return sqrt((data*data.conj()).real,out=out_array)
    return abs_value
def _generate_abs_value_cache_aware(out_array,L2_cache):
    data_shape = out_array.shape
    data_type = out_array.dtype
    splitting_dimension,step = get_L2_cache_split_parameters(data_shape,data_type,L2_cache)

    mult = np.multiply
    sqrt = np.sqrt
    
    def abs_value_1_loop(data):
        for i in range(0,data_shape[0],step):
            i2 = i+step
            d = data[i:i2]
            o = out_array[i:i2]
            sqrt(mult(d,d.conj(),out = o).real,out=o)
        return out_array
    def abs_value_2_loop(data):
        for i in range(data_shape[0]):
            for j in range(0,data_shape[1],step):
                j2 = j+step
                d = data[i,j:j2]
                o = out_array[i,j:j2]
                sqrt(mult(d,d.conj(),out = o).real,out=o)
        return out_array
    def abs_value_3_loop(data):
        for i in range(data_shape[0]):
            for j in range(data_shape[1]):
                for k in range(0,data_shape[2],step):
                    k2 = k+step
                    d = data[i,j,k:k2]
                    o = out_array[k:k2]
                    sqrt(mult(d,d.conj(),out = o).real,out=o)
        return out_array
    
    if splitting_dimension <= -1:
        abs_value = _generate_abs_value_default(out_array)
    elif splitting_dimension == 0:
        abs_value = abs_value_1_loop
    elif splitting_dimension == 1:
        abs_value = abs_value_2_loop
    elif splitting_dimension == 2:
        abs_value = abs_value_3_loop
    return abs_value

def generate_absolute_value(data_shape,data_type,cache_aware=False,L2_cache=256):
    out = np.zeros(data_shape,dtype = data_type)
    if not cache_aware:
        abs_value = _generate_abs_value_default(out)
    else:
        abs_value = _generate_abs_value_cache_aware(out,L2_cache)
    return abs_value


def generate_calc_center_old(real_grid):
    dim = real_grid.n_shape[0]
    cart_grid = spherical_to_cartesian(real_grid)    
    def calc_center(density):
        density_sum=np.sum((density.real*real_grid[...,0]).flatten())
        if density_sum==0:
            density_sum=1
        center=np.sum((cart_grid*(real_grid[...,0]*density.real)[...,None]).reshape(-1,dim),axis=0)/density_sum
        log.info('center={}'.format(center))
        center=cartesian_to_spherical(center)
        #polygon=mLib.SampleShapeFunctions.get_disk_function(0.05,amplitude_function=lambda points: np.full(points.shape[:-1],0.01),coordSys=1,center=center)
        #temp_density=np.abs(density)+polygon(real_grid)
        #pres.present(temp_density)
        log.info('center={}'.format(center))
        return center        
    return calc_center

def generate_calc_center(real_grid):
    cart_grid = spherical_to_cartesian(real_grid)
    dim = real_grid[:].shape[-1]

    if dim ==2:
        si = mLib.PolarIntegrator(real_grid[:])
    elif dim == 3 :
        si = mLib.SphericalIntegrator(real_grid[:])
        
    def calc_center(density):            
        density_integral = si.integrate(density.real)            
        if density_integral==0:
            density_integral=1
        center = si.integrate(cart_grid[:]*density[...,None].real)/density_integral
        #log.info('ccenter={}'.format(center))
        center=cartesian_to_spherical(center)
        return center        
    return calc_center


def check_negative_intensity(intensity):
    n_negative_values = np.sum(intensity.real<0)/np.prod(intensity.shape)
    #log.info('negative fraction = {}'.format(n_negative_values))

def diff(a,b):
    return a-b
def add(a,b):
    #mask = (b*b.conj()).real<1e-3*(a*a.conj()).real
    #log.info('masked fraction = {}'.format(np.sum(mask)/np.prod(mask.shape)))
    #log.info('a type = {}, b_type = {}'.format(type(a),type(b)))
    return a+b
def add_above_zero_index(a,b):
    result = a+b
    result[0]=a[0]
    return result
def add_low(a,b):
    mask = (b*b.conj()).real<1e-3*(a*a.conj()).real
    log.info('masked fraction = {}'.format(np.sum(mask)/np.prod(mask.shape)))
    return a #np.where(mask,a+b,a)

def debug(a,b):
    log.info(' {}'.format(a[0][:10]))
    #for aa in a :
    #    aa[:]=0
    a[0][:]=10
    #log.info(' {}'.format(a[0][:10]))
    return a
    #args = [a,b]
    #log.info('Got {} arguments with types {}'.format(len(args),[type(a) for a in args]))
    #for i,a in enumerate(args):
    #    if isinstance(a,np.ndarray):
    #        log.info('Argument {} is array of shape {} and dtype {}'.format(i,a.shape,a.dtype))
    
def project(a):
    log.info('old 0 coeff = {}'.format(a[0][:10]))
    a[0][:]=1.9e1
    log.info('new 0 coeff = {}'.format(a[0][:10]))
    return a

def project2(new_I,I,d):
    #log.info('first new I = {}'.format(new_I[1,1,:10]))
    #log.info('first old I = {}'.format(I[1,1,:10]))
    n = d/np.sqrt(I)
    n *= np.sqrt(new_I)
    return n

def sqrt2(a):
    #log.info(a)
    aa = np.sqrt(a)
    b = aa*aa.conj()
    return b

def check(a):
    if np.isinf(a).any():
        log.info('inf values found')
    if np.isnan(a).any():
        log.info('nan values found')

def check3(a):
    if np.isinf(a).any():
        log.info('inf values found 3')
    if np.isnan(a).any():
        log.info('nan values found 3')

def check2(Ilm):
    for I in Ilm:
        if np.isinf(I).any():
            log.info('inf values found 2')
        if np.isnan(I).any():
            log.info('nan values found 2')
            

def _get_reciprocity_coefficient(ft_opt):
    pi_in_q = ft_opt.get('pi_in_q',None)
    
    if isinstance(pi_in_q,bool):
        reciprocity_coefficient = _pi_in_q__to__reciprocity_coefficient(pi_in_q)
    else:
        reciprocity_coefficient = ft_opt.get('reciprocity_coefficient',np.pi)
    return reciprocity_coefficient




from xframe.analysis.analysisWorker import AnalysisWorker
from xframe import Multiprocessing
import numpy as np

def multiply_matrix_with_vectors(vects,matrix,**kwargs):
    vects= np.atleast_2d(vects)
    new_vects = np.sum(matrix[None,:,:]*vects[:,None,:],axis=2)
    return np.squeeze(new_vects)
	
class Worker(AnalysisWorker):
    def start_working(self):
        vectors = np.random.rand(200,10)
        matrix = np.random.rand(10,10)
        result = Multiprocessing.process_mp_request(multiply_matrix_with_vectors,input_arrays=[vectors],const_inputs = [matrix])
        test_result = multiply_matrix_with_vectors(vectors,matrix)
        if (result == test_result).all():
            print('Test passed!')




import numpy as np
import os
import glob
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi,heat2D,heatPolar2D,agipd_heat
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward,cartesian_to_spherical,spherical_to_cartesian
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward,masked_mean
from xframe.presenters.matplolibPresenter import plot1D,hist2D,heat2D,scatter1D



import logging
#from xframe.plugins.fxs3046_online.analysisLibrary.regrid2 import SimpleRegridder2D,AgipdRegridderSimple

from scipy.spatial.kdtree import KDTree
from scipy.ndimage import gaussian_filter

log=logging.getLogger('root')

init = False
if init:
    from xframe.startup_routines import dependency_injection
    dependency_injection()
    from xframe.startup_routines import load_recipes
    from xframe.control.Control import Controller
    analysis_sketch,exp_sketch = load_recipes(ana_name='fxs3046_online.generate_plots_2', ana_settings='saxs_65_-bg',exp_name='fxs3046_online',exp_settings = 'default_proc')
    controller = Controller(analysis_sketch,experiment_sketch = exp_sketch)
    controller.chooseJobSingleProcess()
    from xframe import database
    from xframe import Multiprocessing
    experiment = Multiprocessing.comm_module.get_experiment()
    aw = controller.controlWorker.analysis_worker
    aw.geometry = aw.load_geometry()
    aw.agipd_regridder = False
    r = controller.controlWorker.analysis_worker.set_and_get_agipd_regridder()

db = default_DB()
#base_path = '/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_65/'
base_path = '/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_140/'
base_path = '/gpfs/exfel/theory_group/user/berberic/p3046/data/fig/run_{run}/'
#time = '2022-09-03_19:22:42/bg_data_proc.h5'
#time = '2022-09-03_20:15:09/bg_data_proc.h5'
#time = '2022-09-03_21:53:26/bg_data_proc.h5' #run 65 100k
#time = '2022-09-03_23:49:52/bg_data_proc.h5'
#time = '2022-09-04_00:11:48/bg_data_proc.h5' # run 140 100k 0.1,0.3
#time = '2022-09-04_00:11:49/bg_data_proc.h5' # run 140 100k 1.9-2.1
#time = '2022-09-04_12:05:46/pump_diff_run_140_random.h5' #run 140 10k random
#time = '2022-09-04_13:20:07/pump_diff_run_140_random.h5' #run 140 100k random
time = '2022-09-04_10:54:36/bg_data_proc.h5' #run 140 all patterns 
#d = db.load(base_path +time+'/'+ 'bg_data_proc.h5')
time = '2022-09-04_21:04:54/pump_diff_run_140_random.h5' #run 168 all patterns 
d = db.load(base_path.format(run = 168) +time)

d['polar_grid']=r.new_grid
db.save(base_path +time,d)

patterns = d['diod_diff']['saxs_patterns']
#qs = np.arange(512,dtype = float) #d['diod_diff']['polar_grid'][:,0,0]
qs = r.new_grid[:,0,0]

q_mask_max = qs>0.15
q_mask_min = qs<2.5
q_mask1 = q_mask_min & q_mask_max

q_mask_max = qs>0.15
q_mask_min = qs<0.5
q_mask2 = q_mask_min & q_mask_max

q_mask_max = qs > 1.9
q_mask_min = qs < 2.5
q_mask3 = q_mask_max & q_mask_min

q_mask_max = qs > 0.12
q_mask_min = qs < 2
q_mask4 = q_mask_max & q_mask_min


#q_mask_max = qs > 0.12
q_mask_min = qs < 2
q_mask5 =  q_mask_min


patterns_c = []
cells = np.array(tuple( int(c) for c in patterns))
cell_sort = np.argsort(cells)
sorted_cells = cells[cell_sort]
cell_list = []
for key in sorted_cells:
    patterns_c += list(patterns[str(key)]['saxs'])
    #u,s,vh = np.linalg.svd(patterns[str(key)]['saxs'][:,q_mask4],full_matrices=False,compute_uv = True)
    #fig = plot1D.get_fig(value['saxs'],grid = qs)
    #db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/c_{}.matplotlib'.format(key),fig)
    cell_list.append([key]*len(patterns[str(key)]['saxs']))
patterns_c = np.array(patterns_c)
u,s,vh = np.linalg.svd(patterns_c[:,q_mask4],full_matrices=False,compute_uv = True)
cell_ids = np.concatenate(cell_list)

cell_wise_mean_weight_components = []
cell_wise_saxs_no_0 = []
cell_wise_saxs = []
for i in sorted_cells:
    cell_mask = np.nonzero(cell_ids==i)
    mean_weight_components = np.mean(u[cell_mask],axis = 0)
    cell_wise_mean_weight_components.append(mean_weight_components)
    saxs_no_0 = np.mean((u[:,1:]*s[1:]@vh[1:])[cell_mask],axis = 0)
    saxs = np.mean(patterns_c[cell_mask],axis = 0)
    cell_wise_saxs_no_0.append(saxs_no_0)
    cell_wise_saxs.append(saxs)
cws = np.array(cell_wise_saxs)
cws_no0 = np.array(cell_wise_saxs_no_0)
cww =np.array(cell_wise_mean_weight_components)
cww = cww.T

saxs = np.mean(cws,axis = 0)
    
log.info('n_patterns = {}'.format(patterns_c.shape))
chunksize = 4

#new_dataset = np.zeros((patterns_c.shape[0],)+(patterns_c.shape[1]//chunksize,))
#for i in range(0,patterns_c.shape[-1],chunksize):
#    new_dataset[:,i//4] = np.mean(patterns_c[:,i:i+chunksize],axis = -1)
#new_qs = np.zeros_like(new_dataset)
#new_qs[:]= qs[2::chunksize][None,:]

means = np.mean(patterns_c[:,q_mask2],axis=1)
filter_mask1 =  np.abs(means)<1e-1
filter_mask2 =  np.abs(means)>1e-6
filter_mask3 =  np.all(patterns_c[:,q_mask3] <0,axis=1)
filter_mask = filter_mask1 & filter_mask2 #& filter_mask3
print(patterns_c.shape)
patterns_c=patterns_c[filter_mask]
print(patterns_c.shape)

stacked_qs = np.zeros_like(patterns_c)
stacked_qs[:] = qs[None,:]


hist = np.stack((stacked_qs[:,q_mask1],patterns_c[:,q_mask1]),axis =-1).reshape(-1,2)

bins = np.sum(q_mask1)
#fig = hist2D.get_fig(hist,bins=bins,range=[[hist[...,0].min(),hist[...,0].max()],[-5e-1,5e-1]],norm='log',layout={'title':'Difference SAXS/WAXS histogram Run 140 \n normalization region [0.1,0.3] $A^{-1}$ ,randomized pump','x_label': "q [$A^-1$]",'y_label':'I [arb.]'})
#
#fig2 = plot1D.get_fig(np.mean(patterns_c,axis = 0)[qs<3],grid = qs[qs<3],ylim=[-0.005,0.005],layout={'title':'Difference SAXS/WAXS Run 140 \n normalization region [0.1,0.3] $A^-1$ randomized pump','x_label': "q [$A^{-1}$]",'y_label':'I [arb.]'})
#h = np.histogram2d(hist[...,0],hist[...,1],range=[[hist[...,0].min(),hist[...,0].max()],[-0.0005,0.0005]])[0]#,bins=np.sum(q_mask),)[0]

#


#cell_wise_svd = []
#cell_wise_no_0_component=[]
#cell_wise_no_0_component_saxs=[]
#cell_wise_single_weight = []
#for key,value in patterns.items():
#    u,s,vh = np.linalg.svd(value['saxs'][:,q_mask4],full_matrices=False,compute_uv = True)
#    cell_wise_svd.append([u,s,vh])
#    no_zero = u[:,1:]*s[1:]@vh[1:]
#    cell_wise_no_0_component.append(no_zero)
#    cell_wise_no_0_component_saxs.append(np.mean(no_zero,axis=0))
#    #ssvd_component = np.zeros((100,vh.shape[-1]))
#    #for i in range(100):
#    #    ssvd_component[i] = np.mean(u[:,i,None]*s[i]*vh[None,i],axis = 0)
#    weight_mean = np.mean(u,axis = 0)[:100]
#    cell_wise_single_weight.append(weight_mean)
        
    #fig = plot1D.get_fig(value['saxs'],grid = qs)
    #db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/c_{}.matplotlib'.format(key),fig)

#s =db.load('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/svd_s.npy')
#u =db.load('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/svd_u.npy')
#vh =db.load('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/svd_vh.npy')
#
m = np.diag(s)@vh

fig2 = scatter1D.get_fig(s[:500],s=3,layout={'title':'svd of pulse/Difference SAXS/WAXS Run 140 \n normalization region [0.1,0.3] $A^-1$ randomized pump','x_label': "Singular values ",'y_label':'magnitude'},y_scale='log')

fig2 = plot1D.get_fig((m[:10]+np.arange(1,11,dtype = float)[:,None]*50)[::-1],labels=list(np.arange(10).astype(str))[::-1],grid = qs[q_mask4],layout={'title':'s@vh svd of pulse/Difference SAXS/WAXS Run 140 \n normalization region [0.1,0.3] $A^-1$ randomized pump','x_label': "q [$A^{-1}$]",'y_label':'I [arb.]'},y_scale='log')

#fig2 = scatter1D.get_fig(s[:500],s=0.5,layout={'title':'svd of pulse/Difference SAXS/WAXS Run 140 \n normalization region [0.1,0.3] $A^-1$ randomized pump','x_label': "Singular values ",'y_label':'magnitude'},y_scale='log')e('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/svd.matplotlib'.format(key),fig2)

#fig2 = plot1D.get_fig(a,layout={'title':'Singular values of pulse/Difference SAXS/WAXS Run 140 \n normalization region [0.1,0.3] $A^-1$ randomized pump','x_label': "q [$A^{-1}$]",'y_label':'I [arb.]'})
#fig3 = heat2D.get_fig(h,scale='log')
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/r140_hist_wide_random.matplotlib'.format(key),fig)

#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/r140_hist_wide_all_4.matplotlib'.format(key),fig)
#db.save('/gpfs/exfel/theory_group/user/berberic/p3046/pump_test/mean_mp_test_all_4.matplotlib'.format(key),fig2)


#data = d['diod_diff']['mean_on_polar_grid']
#mask = d['diod_diff']['mask_on_polar_grid'].astype(float)
#grid = GridFactory.construct_grid('uniform',[np.arange(data.shape[0],dtype=float),np.arange(data.shape[1],dtype=float)*2*np.pi/data.shape[1]])[:]
#db.save(base_path +time+'/mean.vts',[data],grid = grid,dset_names=['s','t'],grid_type=['polar'],)

#saxs,counts = masked_mean(data,mask,axis=1)



import numpy as np
import logging
log = logging.getLogger("root")
from xframe.library.interfaces import SoftInterface
import pysofft
from pysofft.make_wiegner import CosEvalPts,CosEvalPts2,SinEvalPts,SinEvalPts2,genWigTrans_L2
from pysofft.make_wiegner import genWigAll,genWigAllTrans,get_euler_angles
from pysofft.wignerTransform import wigNaiveSynthesis_fftw
from pysofft.wignerWeights import makeweights2
from pysofft.soft import Inverse_SO3_Naive_fft_pc,Forward_SO3_Naive_fft_pc,coefLoc_so3,sampLoc_so3,totalCoeffs_so3
from pysofft.soft import sampLoc_so3, calc_mean_C_array
from pysofft.rotate import rotate_coeff_multi,rotate_coeff
from xframe.library.gridLibrary import GridFactory
from pysofft import soft as _soft


class Soft(SoftInterface):
    def __init__(self,bw):
        self.bw = bw
        self._soft = _soft
        tmp = self.generate_data()
        self.grid = self.make_SO3_grid()
        self.wigners = tmp[0]
        self.wigners_transposed = tmp[1] #small d matrices with m1,m2 swapped
        self.legendre_weights = tmp[2]
        self.n_coeffs = totalCoeffs_so3(bw)
        self.n_points = (2*bw)**3
        
    def generate_data(self):
        band_width = self.bw
        wigners = genWigAll(band_width)
        wigners_transposed = genWigAllTrans(band_width)
        legendre_weights = makeweights2(band_width)
        return wigners, wigners_transposed, legendre_weights
    
    def forward_cmplx(self,data):        
        if data.ndim>1:
            data=data.flatten()
        if (data.dtype != np.dtype(np.complex128)):
            log.warning('input dtype is {} but {} is required. Trying to change dtype.'.format(data.dtype,np.dtype(np.complex128)))
            data = data.astype(np.complex128)
        if (len(data) != self.n_points):
            raise AssertionError('input data needs to be of length {}. Given length = {}'.format(self.n_points,data.dtype,len(data)))        
        coeffs = Forward_SO3_Naive_fft_pc(self.bw,data,self.legendre_weights,self.wigners,True)        
        return coeffs
    
    def inverse_cmplx(self,coeff):
        if (coeff.dtype != np.dtype(np.complex128)):
            log.warning('input dtype is {} but {} is required. Trying to change dtype.'.format(coeff.dtype,np.dtype(np.complex128)))
            coeff = coeff.astype(np.complex128)
        if (len(coeff) != self.n_coeffs) or (coeff.dtype != np.dtype(np.complex128)):
            raise AssertionError('input coeffs need to be of type complex128 and of length {}. Given data dtype = {} length = {}'.format(self.n_coeffs,coeff.dtype,len(coeff)))        
        data = Inverse_SO3_Naive_fft_pc(self.bw,coeff,self.wigners_transposed,True)        
        return data

    def make_SO3_grid(self):
        alpha,beta,gamma = get_euler_angles(self.bw)
        grid = GridFactory.construct_grid('uniform',[alpha,beta,gamma])[:]
        return grid
    

    def get_empty_coeff(self):
        return np.zeros(self.n_coeffs,dtype = complex)
        


    #################################################
    # computes the rotated harmonic coefficients
    # f_{l,m} ---> \sum_n D^l_{n,m} f_{l,n}
    # assumes f_{l,m}=coeff to be a complex array of length bw**2
    # split_ids such that np.split(coeff,split_ids) gives a list of coefficients indexed by l.
    # euler_angles is an array of length 3 containing the euler angles (alpha,beta,gamma) in ZYZ format
    def rotate_coeff(self,coeff,split_ids,euler_angles):
        #log.info('coeff.shape before rotation = {}'.format(coeff[0].shape))
        rotated_coeff = rotate_coeff_multi(self.bw, coeff,split_ids, euler_angles)
        return rotated_coeff

    def rotate_coeff_single(self,coeff,split_ids,euler_angles):
        log.info('coeff.shape before rotation = {}'.format(coeff.shape))
        rotated_coeff = rotate_coeff(self.bw, coeff,split_ids, euler_angles)
        return rotated_coeff

    
    #######################################
    # let f,g be two square integrable functions on the 2 sphere 
    # Define C: SO(3) ---> R,   R |---> <f,g \circ R> = \int_{S^2} dx f(x)*\overline{g(Rx)}
    # This function calculates C(R) for all R defined in make_SO3_grid
    #
    # arguments :
    #   f_coeff: f_{l,m} spherical harmonic coefficients of f 
    #   g_coeff: g_{l,m} spherical harmonic coefficients of g 
    #            f_coeff, g_coeff are complex numpy arrays of shape bw*(bw+1)+bw+1
    #   split _ids: ids that split coefficients in 2*bw+1 sub arrays indexed by m
    #
    def calc_mean_C(self,f_coeff,g_coeff,r_split_ids,ml_split_ids):
        r_split_lower=r_split_ids[0]
        r_split_upper=r_split_ids[1]
        mean_C = calc_mean_C_array(self.bw,f_coeff,g_coeff,r_split_lower,r_split_upper,ml_split_ids,self.wigners_transposed,True)
        return mean_C


    #testing 
    def combine_coeffs(self,f_coeff,g_coeff,split_ids):
        return combine_harmonic_coeffs(self.bw,f_coeff,g_coeff,split_ids)



import logging
log = logging.getLogger('root')
from flint import arb,arb_mat
from flint import acb,acb_mat
from flint import good
import numpy as np
from xframe.library.interfaces import GSLInterface
import time
_jacobi_p = arb.jacobi_p
_bessel_J = arb.bessel_j
_sqrt = arb.sqrt
_pi = arb.pi
half = arb('1/2')

class ARBPlugin(GSLInterface):
    lib = arb
    good = good    
    @staticmethod
    def jacobi_p(x,n,a,b):        
        def generate_functions(x,n,a,b):
            def f():
                return _jacobi_p(arb(float(x)),int(n),float(a),float(b))
            return f
        f_list=tuple(map(generate_functions,x,n,a,b))
        
        results = np.array(tuple(float(good(f,dps=15,maxprec=100000).str(radius=False)) for f in f_list))
        return results

    @staticmethod
    def spherical_bessel_arb(x,n):
        value = _sqrt(_pi()/(2*x))*_bessel_J(x,n+half)
        if x==0:
            if n==0:
                value=arb(1)
            else:
                value=arb(0)
        return value
    @staticmethod
    def zernike_ND_arb(m,n,x,D=2):
        nm2 = int((n-m)/2)
        return (-1)**(nm2)*x**m*_jacobi_p(1-2*x**2,nm2,m+D/2-1,0)
    @staticmethod
    def zernike_ND_arb_no_sign(m,n,x,D=2):
        return x**m*_jacobi_p(1-2*x**2,int((n-m)/2),m+D/2-1,0)
    
    @staticmethod
    def zernike_ND(m,n,x,D=2):
        D = arb(D)
        #((-1)**((n-m)/2))*(points**m)*eval_jacobi((n-m)/2,m+D/2-1,0,1-2*(points**2))
        def generate_functions(m,n,x_val):
            def f():
                x = arb(float(x_val))
                nm2 = int((n-m)/2)
                return (-1)**(nm2)*x**m*_jacobi_p(1-2*x**2,nm2,int(m)+D/2-1,0)
            return f
        f_list=tuple(map(generate_functions,m,n,x))
        
        results = np.array(tuple(float(good(f,dps=15,maxprec=100000).str(radius=False)) for f in f_list))
        return results
    @staticmethod
    def zernike_ND2(m,n_max,x,D=2):
        print(D)
        D = arb(D)
        n_points = len(x)
        ms = [m]*n_points
        result=[]
        zernike_ND = ARBPlugin.zernike_ND
        for n in range(m,n_max,2):
            print(n)
            result.append(zernike_ND(ms,[n]*n_points,x))
        return np.array(result)

    @staticmethod
    def calc_spherical_zernike_wights_fixed_ord_pi_worker(k,p,order,n_radial_points,expansion_limit,**kwargs):
        ''' k >= 1 , p>=0'''
        R_ND = ARBPlugin.zernike_ND_arb_no_sign
        jnu = ARBPlugin.spherical_bessel_arb
        #radial_points=opt['n_radial_points']+1        
        if p == 0:
            k2p = k**2
        else:
            k2p = k**2/p
        #k2p = p_not_zero_case*(k**2/p)+p_zero_case*(k**2) 
        expansion_range=range(order,expansion_limit+1,2)
        #print('k2p={}'.format(k2p))
        #print('R_ND={}'.format(tuple(R_ND(order,n,k/n_radial_points) for n in expansion_range)))

        summands = tuple((2*n+3)*R_ND(0,n,k/n_radial_points,D=3)*jnu(p*_pi(),n+1) for n in expansion_range)
        if (p==0) and (order==0):
            summands= ((2*0+3)*R_ND(order,0,k/n_radial_points,D=3),) + summands[1:]
        weight = sum(summands)
        return weight*k2p*_sqrt(2/_pi())
    
    @staticmethod
    def calc_spherical_zernike_wights_fixed_ord_pi(order,n_radial_points,expansion_limit,**kwargs):
        worker = ARBPlugin.calc_spherical_zernike_wights_fixed_ord_pi_worker
        ks = tuple(arb(int(k)) for k in np.arange(1,n_radial_points))
        ps = tuple(arb(int(p)) for p in np.arange(n_radial_points))
        start = time.time()
        def gen_functions(k,p):
            if p==0:
                print( time.time() - start )
            def f():
                return worker(k,p,order,n_radial_points,expansion_limit)
            return f
        
        weights=np.array(
            tuple(
                tuple(
                    float(good(gen_functions(k,p),dps=15,maxprec=100000).str(radius=False)) for p in ps) for k in ks)
        )
        return weights

        
        



import numpy as np
import numpy.ma as mp
import scipy.integrate as spIntegrate
import logging
from itertools import repeat

log=logging.getLogger('root')

import xframe.library.pythonLibrary as pyLib
import xframe.library.mathLibrary as mLib
from .harmonic_transforms import generate_circular_harmonic_transform_pair
from .harmonic_transforms import HarmonicTransform
from .hankel_transforms import generateHT_SinCos
from .hankel_transforms import generate_zernike_ht
from .hankel_transforms import generate_spherical_zernike_ht
from .hankel_transforms import generate_spherical_zernike_ht_new
from .hankel_transforms import generate_spherical_zernike_ht_new2
from .hankel_transforms import generate_spherical_zernike_ht_big_sum
from .hankel_transforms import generate_spherical_zernike_ht_gpu
from .hankel_transforms import generate_ht_spherical_SinCos
from .hankel_transforms import generate_weightDict_zernike_spherical
from xframe import database
from xframe.settings import analysis as settings

doublePrecision=pyLib.doublePrecision



###@#########################
###polar Fourier via Hankel transforms###
def generatePolarFT_viaHT(FTGridPair,generateHankeltransforms):
    dht_inverse,dht_forwars=generateHankeltransforms(FTGridPair)

    def fourierTransform(dataGrid):
        harmonicCoefficients=circularHarmonicTransform(dataGrid)
#        log.info('harmonicCoefficients shape={}'.format(harmonicCoefficients.shape))
        reciprocalHarmonicCoefficients=dht_forward(harmonicCoefficients,quadratureWeights,realCutOff)
#        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocalFunction=circularHarmonicTransform_inverse(reciprocalHarmonicCoefficients)
        return reciprocalFunction
    
    def inverseFourierTransform(dataGrid):
        reciprocalHarmonicCoefficients=circularHarmonicTransform(dataGrid)
        harmonicCoefficients=dht_inverse(reciprocalHarmonicCoefficients,quadratureWeights,reciprocalCutOff)
        realFunction=circularHarmonicTransform_inverse(harmonicCoefficients)
        return realFunction
    return fourierTransform,inverseFourierTransform

####################
###polar bessel forward####
def polarFT_forward_bessel(sampledPolarFunction):
    polarIndexList=generatePolarIndexList(sampledPolarFunction.shape[-1])
    indexCount=0
    def dht(inputData):
        nonlocal indexCount
#        log.info('indexCount={}'.format(indexCount))
        polarIndex=polarIndexList[indexCount]        
        #      log.info('input data ={}'.format(inputData))
        transformedData=mLib.dht_forward(inputData,polarIndex)
        transformedData=transformedData*2*np.pi*1.j**(-polarIndex)
        indexCount+=1
        return transformedData        
    
    shape=sampledPolarFunction.shape
    fftForward=mLib.fft_real_mixedRadix_forward
    fftInverse=mLib.fft_complex_mixedRadix_inverse
    
#    log.info('\n in=\n {}'.format(sampledPolarFunction.array))
    harmonicCoefficients=sampledPolarFunction.apply_along_axis(fftForward,len(shape)-1)
#    log.info('\n ft=\n {}'.format(harmonicCoefficients.array))
    reciprocalHarmonicCoefficients=harmonicCoefficients.apply_along_axis(dht,0)
#    log.info('\n dht =\n {}'.format(reciprocalHarmonicCoefficients.array))
    reciprocalFunction=reciprocalHarmonicCoefficients.apply_along_axis(fftInverse,len(shape)-1)
#    log.info('after forwart FT =\n {}'.format(reciprocalFunction.array))
    return reciprocalFunction


###################
###polar bessel inverse###
def polarFT_inverse_bessel(dataGrid):    
    polarIndexList=generatePolarIndexList(dataGrid.shape[-1])
    indexCount=0
    def inverseDHT(inputData):
        nonlocal indexCount
#        log.info('indexCount={}'.format(indexCount))
        polarIndex=polarIndexList[indexCount]        
        #log.info('input data ={}'.format(inputData))
        newInputData=inputData*1.j**polarIndex/(2*np.pi)
        transformedData=mLib.dht_inverse(newInputData,polarIndex)
        indexCount+=1
        return transformedData

    shape=dataGrid.shape
    fftForward=mLib.fft_complex_mixedRadix_forward
    fftInverse=mLib.fft_complex_mixedRadix_inverse
#    log.info('\n in=\n {}'.format(dataGrid.array))
    reciprocalHarmonicCoefficients=dataGrid.apply_along_axis(fftForward,len(shape)-1)
#    log.info('\n ft=\n {}'.format(reciprocalHarmonicCoefficients.array))
    harmonicCoefficients=reciprocalHarmonicCoefficients.apply_along_axis(inverseDHT,0)
#    log.info('\n dht^-1=\n {}'.format(harmonicCoefficients.array))
    function=harmonicCoefficients.apply_along_axis(fftInverse,len(shape)-1)
#    log.info('after inverse FT =\n {}'.format(function.array))
    return function



##################
### adaptiveQuad FTs ###
def generateAdaptive_polarFT(FTGridPair):
    #assumes FTGridPair was generated by the polarFTGridPair_Donatelli  method
    realGrid=FTGridPair.realGrid
    reciprocalGrid=FTGridPair.reciprocalGrid
    realOrigin=realGrid.flatten()[0]
    reciprocalOrigin=reciprocalGrid.flatten()[0]    
    gridDimension=len(realOrigin)

    realStepSizes=pyLib.uniformGridGetStepSizes(realGrid.array)
    reziprocalStepSizes=pyLib.uniformGridGetStepSizes(reziprocalGrid.array)
    realGridSpec=[realOrigin,realStepSizes]
    reziprocalGridSpec=[reziprocalOrigin,reziprocalStepSizes]
    
    def fourierTransform(dataGrid):
        harmonicCoefficients=circularHarmonicTransform(dataGrid)
        reciprocalHarmonicCoefficients=hankelTransform(harmonicCoefficients,realGridSpec,realCutOff)
        

###############
### donatelli FTs ###
def generate_discrete_polarFT(maxR,n_radial_points,harmonic_orders):
    dht_forward,dht_inverse=mLib.generate_dht_2d(maxR,n_radial_points,harmonic_orders)
    cht_forward,cht_inverse=generate_circular_harmonic_transform_pair('complex',len(harmonic_orders))

    def ft_forward(real_function):
        return cht_inverse(dht_forward(cht_forward(real_function)))
    def ft_inverse(reciprocal_function):
        return cht_inverse(dht_inverse(cht_forward(reciprocal_function)))
    return ft_forward,ft_inverse

def generateDonatelli_polarFT_2(ft_grid_pair,weights_dict,mode):
    real_grid=ft_grid_pair.realGrid
    
    if mode=='real':
        cht_forward,cht_inverse=generate_circular_harmonic_transform_pair('real',real_grid.shape[-1])
        ht_forward,ht_inverse=generateHT_SinCos(ft_grid_pair,weights_dict,mode='real')
    elif mode=='complex':
        cht_forward,cht_inverse=generate_circular_harmonic_transform_pair('complex',real_grid.shape[-1])
        ht_forward,ht_inverse=generateHT_SinCos(ft_grid_pair,weights_dict,mode='complex')
        
        
        #    step_sizes=pyLib.uniformGridGetStepSizes(real_grid.array)
        
    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=cht_forward(data)
        #        log.info('harmonicCoefficients shape={}'.format(harmonicCoefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=ht_forward(harmonic_coefficients)
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=cht_inverse(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=cht_forward(data)
        harmonic_coefficients=ht_inverse(reciprocal_harmonic_coefficients)
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=cht_inverse(harmonic_coefficients)
        return real_function
    return fourierTransform,inverseFourierTransform

def generateDonatelli_polarFT(ft_grid_pair,weights_dict,mode):
    real_grid=ft_grid_pair.realGrid
    n_angular_points=real_grid.shape[-1]
    n_independent_coeff=int(np.floor((n_angular_points+1)/2)+n_angular_points%2)
    ht,iht=generateHT_SinCos(ft_grid_pair,weights_dict)
    complex_cht,complex_icht=generate_circular_harmonic_transform_pair('complex',real_grid.shape[-1])
    if mode=='complex':
        hankel=ht
        i_hankel=iht
        ft_cht=complex_cht
        ft_icht=complex_icht
        ift_cht=complex_cht
        ift_icht=complex_icht     
    elif mode=='real':
        ft_cht,ift_icht=generate_circular_harmonic_transform_pair('real',real_grid.shape[-1])
        def hankel(harmonic_coefficients):
#            log.info('real harm coef shape={}'.format(harmonic_coefficients.array.shape))
            harmonic_coefficients=np.concatenate((harmonic_coefficients,np.conjugate(harmonic_coefficients[:,n_independent_coeff-(n_angular_points+1)%2:0:-1])),axis=-1)
#            log.info('complex harm coef shape={}'.format(harmonic_coefficients.array.shape))
            reciprocal_harmonic_coefficients=ht(harmonic_coefficients)
            return reciprocal_harmonic_coefficients
        def i_hankel(reciprocal_harmonic_coefficients):
            harmonic_coefficients=iht(reciprocal_harmonic_coefficients)[:,:n_independent_coeff]
            return harmonic_coefficients
        
        def ft_icht(harmonic_coefficients):
            reciprocal_function=complex_icht(harmonic_coefficients)
            zero_to_pi=(reciprocal_function[:,:n_independent_coeff]+reciprocal_function[:,n_independent_coeff:].conjugate())/2
            reciprocal_function[:,:n_independent_coeff]=zero_to_pi
            reciprocal_function[:,n_independent_coeff:]=zero_to_pi.conjugate()
            return reciprocal_function
        ift_cht=complex_cht

#    step_sizes=pyLib.uniformGridGetStepSizes(real_grid.array)

    def fourierTransform(data):
#        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_cht(data)
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #        log.info('harmonicCoefficients shape={}'.format(harmonicCoefficients.shape))
#        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=hankel(harmonic_coefficients)
#        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
#        reciprocalHarmonicCoefficients=harmonicCoefficients
#        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_icht(reciprocal_harmonic_coefficients)
#        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        return reciprocal_function
    
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_cht(data)

        harmonic_coefficients=i_hankel(reciprocal_harmonic_coefficients)
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_icht(harmonic_coefficients)
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_Donatelli_spherical_ft(weights_dict, ht : HarmonicTransform, mode : str):
    pos_orders=weights_dict['posHarmOrders']
    n_radial_points=weights_dict['weights'].shape[-1]
    l_max_zht=int(max(pos_orders))
    m_zht_indices = np.concatenate((np.arange(l_max_zht+1,dtype = int),-np.arange(l_max_zht,0,-1,dtype = int)))
    zht,izht=generate_ht_spherical_SinCos(weights_dict)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['ml']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        m_orders = ht.m
        l_orders = ht.l
        temp_zeros=tuple(np.zeros((n_radial_points,l_max_zht-np.abs(m)+1),dtype=np.complex) for m in m_orders)
    elif mode=='real':
        log.error('real ft using SinCos_3D not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data)
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        temp_coeffs=zht(harmonic_coefficients)
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        temp_coeffs=izht(reciprocal_harmonic_coefficients)
        
        harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    return fourierTransform,inverseFourierTransform

###@##################
###  Zernike FTs  ###
def generate_zernike_polar_ft(ft_grid_pair,weights_dict,harm_trf,mode):
    real_grid=ft_grid_pair.realGrid
    n_radial_points=real_grid.shape[0]+1
    maxQ=real_grid[-1,0,0]
    maxR=n_radial_points/maxQ
    log.info('maxR={} maxQ={} n_radial_points={}'.format(maxR,maxQ,n_radial_points))
    n_angular_points=real_grid.shape[-1]
    n_independent_coeff=int(np.floor((n_angular_points+1)/2)+n_angular_points%2)
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    zht,izht=generate_zernike_ht(weights,pos_orders,maxR)
    complex_cht,complex_icht=harm_trf
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_cht=complex_cht
        ft_icht=complex_icht
        ift_cht=complex_cht
        ift_icht=complex_icht
        temp_zeros=np.zeros(real_grid.shape,dtype=np.complex)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')
        
    if pos_orders[0]==0:
        orders=np.concatenate((pos_orders,-1*pos_orders[:0:-1]))
    else:
        orders=np.concatenate((pos_orders,-1*pos_orders[::-1]))
        
    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_cht(data)[:,orders]
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=temp_zeros.copy()
        reciprocal_harmonic_coefficients[:,orders]=hankel(harmonic_coefficients)
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_icht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_cht(data)[:,orders]
        harmonic_coefficients=temp_zeros.copy()
        harmonic_coefficients[:,orders]=i_hankel(reciprocal_harmonic_coefficients)
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_icht(harmonic_coefficients)
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_zernike_spherical_ft_old(maxR,weights_dict, ht : HarmonicTransform, mode : str):   
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max=int(max(pos_orders))
    zht,izht=generate_spherical_zernike_ht(weights,pos_orders,maxR)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['ml']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        temp_zeros=np.zeros((weights.shape[1],l_max+1),dtype=np.complex)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def apply_hankel(data):
        tmp=temp_zeros.copy()
        tmp[:,orders]=hankel(data[:,orders])
        return tmp
    
    def apply_inverse_hankel(data):
        tmp=temp_zeros.copy()
        tmp[:,orders]=i_hankel(data[:,orders])
        return tmp
        
    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data)
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=list(map(apply_hankel,harmonic_coefficients))
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)        
        harmonic_coefficients=list(map(apply_inverse_hankel,reciprocal_harmonic_coefficients))
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_zernike_spherical_ft_new(maxR,weights_dict, ht : HarmonicTransform, mode : str):   
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max=int(max(pos_orders))
    zht,izht=generate_spherical_zernike_ht_new(weights,pos_orders,maxR)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['ml']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        temp_zeros=tuple(np.zeros((weights.shape[1],l+1),dtype=np.complex) for l in range(l_max+1))
        m_indices = ht.m_indices
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def apply_hankel(data,l_mask,zeros):
        tmp=zeros.copy()
        log.info('temp_zeros shape = {}'.format(tmp.shape))
        tmp[:,l_mask]=hankel(data[:,l_mask])
        return tmp
    
    def apply_inverse_hankel(data):
        tmp=temp_zeros.copy()
        tmp[:,orders]=i_hankel(data[:,orders])
        return tmp
        
    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data)
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=list(apply_hankel(coeff,l_mask,zeros) for coeff,l_mask,zeros in zip(harmonic_coefficients,l_ids,temp_zeros))
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)        
        harmonic_coefficients=list(map(apply_inverse_hankel,reciprocal_harmonic_coefficients))
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_zernike_spherical_ft_new2(maxR,weights_dict, ht : HarmonicTransform, mode : str):   
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max_zht=int(max(pos_orders))
    m_zht_indices = np.concatenate((np.arange(l_max_zht+1,dtype = int),-np.arange(l_max_zht,0,-1,dtype = int)))
    zht,izht=generate_spherical_zernike_ht_new2(weights,pos_orders,maxR)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['ml']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        m_orders = ht.m
        l_orders = ht.l
        temp_zeros=tuple(np.zeros((weights.shape[2],l_orders.max()-np.abs(m)+1),dtype=np.complex) for m in m_orders)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def populate_zeros(zeros,coeff):
        zeros[:len(coeff)]=coeff        

    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data)
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        temp_coeffs=zht(harmonic_coefficients)
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        temp_coeffs=izht(reciprocal_harmonic_coefficients)
        
        harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_zernike_spherical_ft_big_sum(maxR,weights_dict, ht : HarmonicTransform, mode : str):   
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max_zht=int(max(pos_orders))
    m_zht_indices = np.concatenate((np.arange(l_max_zht+1,dtype = int),-np.arange(l_max_zht,0,-1,dtype = int)))
    zht,izht=generate_spherical_zernike_ht_big_sum(weights,pos_orders,maxR)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['direct']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        m_orders = ht.m
        l_orders = ht.l
        temp_zeros=tuple(np.zeros((weights.shape[2],l_orders.max()-np.abs(m)+1),dtype=np.complex) for m in m_orders)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def populate_zeros(zeros,coeff):
        zeros[:len(coeff)]=coeff        

    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data.astype(complex))
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        reciprocal_harmonic_coefficients=zht(harmonic_coefficients)
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        # reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        harmonic_coefficients=izht(reciprocal_harmonic_coefficients)
        
        # harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    return fourierTransform,inverseFourierTransform

def generate_zernike_spherical_ft_gpu_old(maxR,weights_dict, ht : HarmonicTransform, mode : str,mp_lock):
    log.info('use_gpu')
    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max_zht=int(max(pos_orders))
    m_zht_indices = np.concatenate((np.arange(l_max_zht+1,dtype = int),-np.arange(l_max_zht,0,-1,dtype = int)))
    zht,izht=generate_spherical_zernike_ht_gpu(weights,pos_orders,maxR)    
    #grid_shape=(weights.shape[1],sh.)

    trfs=ht.transforms_by_indices['direct']
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        m_orders = ht.m
        l_orders = ht.l
        temp_zeros=tuple(np.zeros((weights.shape[2],l_orders.max()-np.abs(m)+1),dtype=np.complex) for m in m_orders)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)

    def populate_zeros(zeros,coeff):
        zeros[:len(coeff)]=coeff        

    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data.astype(complex))
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        #        with mp_lock: #only one process at a time will call the GPU
        reciprocal_harmonic_coefficients=zht(harmonic_coefficients)            
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        # reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        #        with mp_lock: #only one process at a time will call the GPU
        harmonic_coefficients=izht(reciprocal_harmonic_coefficients)
        
        # harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    return fourierTransform,inverseFourierTransform


def generate_zernike_spherical_ft(maxR,weights_dict, ht : HarmonicTransform, mode : str, pi_in_q : bool, use_gpu = False):

    weights=weights_dict['weights']
    pos_orders=weights_dict['posHarmOrders']
    l_max_zht=int(max(pos_orders))
    m_zht_indices = np.concatenate((np.arange(l_max_zht+1,dtype = int),-np.arange(l_max_zht,0,-1,dtype = int)))
    if use_gpu:
        zht,izht=generate_spherical_zernike_ht_gpu(weights,pos_orders,maxR,pi_in_q = pi_in_q)
        trfs=ht.transforms_by_indices['direct']
    else:
        zht,izht=generate_spherical_zernike_ht(weights,pos_orders,maxR,pi_in_q = pi_in_q)
        trfs=ht.transforms_by_indices['ml']
    #grid_shape=(weights.shape[1],sh.)
    log.info(trfs)
    complex_cht,complex_icht=trfs['forward'], trfs['inverse']
    if mode=='complex':
        hankel=zht
        i_hankel=izht
        ft_sht=complex_cht
        ft_isht=complex_icht
        ift_sht=complex_cht
        ift_isht=complex_icht
        m_orders = ht.m
        l_orders = ht.l
        temp_zeros=tuple(np.zeros((weights.shape[2],l_orders.max()-np.abs(m)+1),dtype=np.complex) for m in m_orders)
    elif mode=='real':
        log.error('real ft using Zernike not implemented at the moment.')

    orders=pos_orders.astype(np.int)
    #log.info('ft_sht = {}'.format(ft_sht))
    def populate_zeros(zeros,coeff):
        zeros[:len(coeff)]=coeff        

    def fourierTransform(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data)
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        temp_coeffs=zht(harmonic_coefficients)
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        temp_coeffs=izht(reciprocal_harmonic_coefficients)
        
        harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        for m in m_zht_indices:
            #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
            coeff=temp_coeffs[m]
            harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function

    def fourierTransform_gpu(data):
        #        log.info('grid dim before fourier={}'.format(dataGrid.dimension))
        harmonic_coefficients=ft_sht(data.astype(complex))
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic shape ={}'.format(harmonic_coefficients[0].shape))
        #log.info('harmonic len ={}'.format(len(harmonic_coefficients)))
        #log.info('harmonic coeff size={}'.format(harmonic_coefficients.array.shape))
        #log.info('harmonic coeff dtype={}'.format(harmonic_coefficients.array.dtype))
        #log.info('harmonicCoefficients shape={}'.format(harmonic_coefficients.shape))
        #        log.info('harmCoefficient dim={}'.format(harmonicCoefficients.dimension))
        #        with mp_lock: #only one process at a time will call the GPU
        reciprocal_harmonic_coefficients=zht(harmonic_coefficients)
        #log.info('reziprocal harmonic shape ={}'.format(reciprocal_harmonic_coefficients[0].shape))
        #log.info('reziprocal harmonic len ={}'.format(len(reciprocal_harmonic_coefficients)))
        #log.info('coeffs shape = {}'.format(tuple(temp_coeffs[m].shape for m in m_zht_indices )))
        # reciprocal_harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     reciprocal_harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        #        log.info('dim after HT={}'.format(reciprocalHarmonicCoefficients.dimension))
        #        reciprocalHarmonicCoefficients=harmonicCoefficients
        #        log.info('reciprocal harmonic coeff = {}'.format(reciprocalHarmonicCoefficients))
        reciprocal_function=ft_isht(reciprocal_harmonic_coefficients)
        #        log.info('dim after FT={}'.format(reciprocalFunction.dimension))
        #log.info('fourier transform done!')
        
        #assuming q_0 = 0 F(0,theta,phi) needs to be constant
        reciprocal_function[0] = reciprocal_function[0].mean()
        return reciprocal_function
    def inverseFourierTransform_gpu(data):
        reciprocal_harmonic_coefficients=ift_sht(data)

        #        with mp_lock: #only one process at a time will call the GPU
        harmonic_coefficients=izht(reciprocal_harmonic_coefficients)
        
        # harmonic_coefficients = tuple(z.copy() for z in temp_zeros)
        # for m in m_zht_indices:
        #     #log.info('m_zht = {} zeros shape = {}'.format(m,temp_zeros[m].shape))
        #     coeff=temp_coeffs[m]
        #     harmonic_coefficients[m][:,:coeff.shape[1]]=coeff
        
        #        harmonicCoefficients=reciprocalHarmonicCoefficients
        real_function=ift_isht(harmonic_coefficients)
        #assuming r_0 = 0 f(0,theta,phi) needs to be constant
        real_function[0] = real_function[0].mean()
        return real_function
    if use_gpu:
        ft = fourierTransform_gpu
        ift = inverseFourierTransform_gpu
    else:
        ft = fourierTransform
        ift = inverseFourierTransform_gpu
    return ft,ift


def zernike_radial_cutoff_relation(cutoff:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal cutoff.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.    
    $2*R*2*Q = 2*N$
    '''
    if pi_in_q:
        other_cutoff = np.pi*n_points/cutoff
    else:
        other_cutoff = n_points/(2*cutoff)
    return other_cutoff
    
def zernike_radial_step_relation(step:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal step sizes.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.
    $\delta R*\delta Q = 1/(2*N)$    
    '''
    if pi_in_q:
        other_step = np.pi*1/(n_points*step)
    else:
        other_step = 1/(2*n_points*step)
    return other_step
    
def zernike_radial_step_cutoff_relation(step:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal step sizes.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.
    $\delta R*\delta Q = 1/(2*N)$    
    '''
    if pi_in_q:
        other_cutoff = np.pi/step
    else:
        other_cutoff = 1/(2*step)
    return other_cutoff
    

################# general selection logic ##################
'''
The following method depends on runtime the runtime variables:
xframe.database.analysis
'''
def generate_fourier_transforms(grid_pair, opt, harm_trf):
    dimension = opt['dimensions']
    ft_type = opt['type']
    data_type = opt['data_type']
    n_radial_points = grid_pair.reciprocal.shape[0]
    harmonic_orders=opt['pos_orders']
    maxOrder=np.max(harmonic_orders)
    nOrders=len(harmonic_orders)
    db = database.analysis
    if dimension == 2:
        if ft_type=='Zernike':
            name_postfix = 'N'+str(n_radial_points)+'mO'+str(maxOrder)+'nO'+str(nOrders)
            weights_dict = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimension)+'D'})
            fourierTransform,inverseFourierTransform = generate_zernike_polar_ft(grid_pair,weights_dict,harm_trf,data_type)
        else:
            log.error("foruier mode {} not known. Replace ourier Transfomrs by identity map.")
            fourierTransform=mLib.id
            inverseFourierTransform=mLib.id
    if dimension==3:
        max_R = grid_pair.real[-1,0,0,0]
        if ft_type=='Zernike':
            pi_in_q = opt['pi_in_q']
            use_gpu = opt['use_GPU']
            if pi_in_q:
                name_postfix='N'+str(n_radial_points)+'mO'+str(maxOrder)+'nO'+str(nOrders)+'_pi'
            else:
                name_postfix='N'+str(n_radial_points)+'mO'+str(maxOrder)+'nO'+str(nOrders)
            try:
                weights_dict = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimension)+'D'})
            except FileNotFoundError as e:
                if opt['allow_weight_calculation']:
                    weights_dict = generate_weightDict_zernike_spherical(maxOrder,n_radial_points,n_cpus=opt.get('n_cpus_weight_generation',False))
                    if opt['allow_weight_saving']:
                        db.save('ft_weights',weights_dict,path_modifiers={'postfix':name_postfix,'type':ft_type+'_'+str(dimension)+'D'})
                                         
            fourierTransform,inverseFourierTransform=generate_zernike_spherical_ft(max_R,weights_dict,harm_trf,data_type,pi_in_q,use_gpu = use_gpu)
           
    return fourierTransform,inverseFourierTransform




#import matplotlib
#matplotlib.use('Agg')
#import matplotlib.pyplot as plt
#matplotlib.rcParams['text.usetex'] = True
#import matplotlib.pyplot as plt
import sys
import numpy as np
import logging
log=logging.getLogger('root')
from xframe.presenters.interfaces import MatplotlibDependency,MPLToolkitDependency


matplotlib = MatplotlibDependency
mpl_toolkits= MPLToolkitDependency

class pltWrapper:
    def __getattr__(self,key):
      return globals()['matplotlib'].pyplot.__getattribute__(key)
  
class locatorWrapper:
    def __getattr__(self,key):
      return globals()['mpl_toolkits'].axes_grid1.inset_locator.__getattribute__(key)
plt = pltWrapper()
inset_locator = locatorWrapper()

def depencency_injection_hook_mpl():
    matplotlib.rcParams['text.usetex'] = True
    module = sys.modules[__name__]
    module.plt = matplotlib.pyplot

def dependency_injection_hook_mpl_toolkits():
    module = sys.modules[__name__]
    module.inset_locator = mpl_toolkits.axes_grid1.inset_locator


from xframe.library import mathLibrary as mLib
from xframe.library.gridLibrary import getGridByXYValues
from xframe.library.gridLibrary import Grid
from xframe.library.pythonLibrary import flattenList
from xframe.library.pythonLibrary import xprint
from xframe.library.pythonLibrary import make_string_tex_conform

def set_axes_equal(ax):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5*max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

    
def apply_layout_to_ax(ax,layout):
    title=layout.get('title',"Scatter plot 2D")
    xLable=layout.get('x_label',"r")
    yLable=layout.get('y_label',"phi")
    title_size = layout.get('title_size',16)
    text_size = layout.get('text_size',14)
    ax.set_title(title,fontsize=title_size)
    ax.set_xlabel(xLable,fontsize=text_size)
    ax.set_ylabel(yLable,fontsize=text_size)
    #ax.legend(ncol=3, loc='best')

def apply_layout_to_figure(fig,layout):
    caption_size = layout.get('caption_size',10)
    caption = layout.get('caption',False)
    if isinstance(caption,str):
        #log.info('caption is added')
        caption = make_string_tex_conform(caption)
        #log.info('caption = {}'.format(caption))
        fig.text(0.1,0.1,caption,fontsize = caption_size,va = 'top',wrap = True)


class Plotter:
    @classmethod
    def get_fig(cls,*args,**kwargs):
        pass
    @classmethod
    def show(cls,*args,**kwargs):
        fig = cls.get_fig(*args,**kwargs)
        fig.show()
class scatter2D:
    def applyLayoutToAx(self,ax,layout):
        title=layout.get('title',"Scatter plot 2D")
        xLable=layout.get('xLable',"x")
        yLable=layout.get('yLable',"y")
        
        ax.set_title(title,fontsize=16)
        ax.set_xlabel(xLable,fontsize=14)
        ax.set_ylabel(yLable,fontsize=14)
    def present2(self):
        pass
    def present(self,DataSets,mode='points',layout={}):
        colors = [[0,0,0]]
        pointArea=1
        for Data in DataSets:
            fig,ax=plt.subplots()
            if mode=='xy':
                x=Data[0]
                y=Data[1]
            else:
                x=Data[:,0]
                y=Data[:,1]
            ax.scatter(x, y,s=pointArea, c=colors, alpha=0.5)

            self.applyLayoutToAx(ax,layout)
#            ax.set_aspect('equal')
        plt.show()


class errorbar2D:
    def applyLayoutToAx(self,ax,layout):
        title=layout.get('title',"Scatter plot 2D")
        xLable=layout.get('xLable',"x")
        yLable=layout.get('yLable',"y")
        
        ax.set_title(title,fontsize=16)
        ax.set_xlabel(xLable,fontsize=14)
        ax.set_ylabel(yLable,fontsize=14)
    def present2(self):
        pass
    def present(self,Data,xerr=None,yerr=None,mode='points',layout={}):
        fig,ax=plt.subplots()
        if mode=='xy':
            x=Data[0]
            y=Data[1]
        else:
            x=Data[:,0]
            y=Data[:,1]
        ax.errorbar(x, y,xerr=xerr,yerr=yerr)

        self.applyLayoutToAx(ax,layout)
#            ax.set_aspect('equal')
        plt.show()

        
class scatterPolar2D:
    def present2(self):
        pass
    def present(self,Data):
        theta=Data[...,1]
        r=Data[...,0]
        colors = [[0,0,0]]
        pointArea=1
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='polar')
        ax.grid(False)
        ax.set_yticklabels([])
        c = ax.scatter(theta, r, c=colors, s=pointArea, alpha=0.5)

        plt.show()
        
class scatter3D:
    def present2(self):
        pass
    def present(self,Data):
        x=Data[:,0]
        y=Data[:,1]
        z=Data[:,2]
        colors = [[0,0,0]]
        pointArea=1
        fig,ax=plt.subplots()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x, y,z,s=pointArea, c=colors, alpha=0.5)
        ax.set_title('Scatter plot 3D')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        plt.show()

class scatterSpherical3D:
    def applyLayoutToAx(self,ax,layout):
        title=layout.get('title',"Scatter plot 3D")
        xLable=layout.get('xLable',"x")
        yLable=layout.get('yLable',"y")
        zLable=layout.get('zLable',"z")
        
        ax.set_title(title,fontsize=16)
        ax.set_xlabel(xLable,fontsize=14)
        ax.set_ylabel(yLable,fontsize=14)
        ax.set_zlabel(zLable,fontsize=14)
        
    def present2(self):
        pass
    def present(self,Data,layout={}):

        cartesianData=mLib.sphericalToCartesian(Data)
        x=cartesianData[:,0]
        y=cartesianData[:,1]
        z=cartesianData[:,2]

        fig=plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        colors = layout.get('colors',[[0,0,0]])
        pointArea=layout.get('pointArea',1)
        alpha=layout.get('alpha',.5)
        ax.scatter(x,y,z,s=pointArea, c=colors, alpha=alpha)

        self.applyLayoutToAx(ax,layout)
        
        
        set_axes_equal(ax)
        plt.show()

class scatterCylindrical3D:
    def applyLayoutToAx(self,ax,layout):
        title=layout.get('title',"Scatter plot 3D")
        xLable=layout.get('xLable',"x")
        yLable=layout.get('yLable',"y")
        zLable=layout.get('zLable',"z")
        
        ax.set_title(title,fontsize=16)
        ax.set_xlabel(xLable,fontsize=14)
        ax.set_ylabel(yLable,fontsize=14)
        ax.set_zlabel(zLable,fontsize=14)
        
    def present2(self):
        pass
    def present(self,Data,layout={}):

        cartesianData=mLib.cylindricalToCartesian(Data)
        x=cartesianData[:,0]
        y=cartesianData[:,1]
        z=cartesianData[:,2]

        fig=plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        colors = layout.get('colors',[[0,0,0]])
        pointArea=layout.get('pointArea',1)
        alpha=layout.get('alpha',.5)
        ax.scatter(x,y,z,s=pointArea, c=colors, alpha=alpha)

        self.applyLayoutToAx(ax,layout)
        
        
        set_axes_equal(ax)
        plt.show()

class hist2D():
    def present(self,data):
        x=data[:,0]
        y=data[:,1]
        fig=plt.figure()
        ax=fig.add_subplot(111)
        ax.hist2d(x,y)
        plt.show()
        
class pcolor2D:
    def present2(self):
        pass
    def present(self,grid,data):
        fig=plt.figure()
        ax=fig.add_subplot(111)
        ax.set_title('Color Mesh')
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_aspect('equal')
        
        x=grid[:,0]
        y=grid[:,1]
        z=data       
        
        c = ax.pcolormesh(x, y, z, cmap='RdBu')
        fig.colorbar(c, ax=ax)

        plt.show()

class heatPolar2D:
    @classmethod   
    def get_fig(cls,data,grid=False,layout={},scale='lin'):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='polar')
        ax.set_yticklabels([])
        ax.set_xticklabels([])

        if not isinstance(grid,bool):
            grid=grid.copy()
            #        log.info('grid shape = {}'.format(grid.shape))
#            grid=np.concatenate((grid[:],np.expand_dims(grid[:,0],1)),axis=1)
            r=grid[...,0]
            phi=grid[...,1]
            #            log.info('phi shape={}'.format(phi.shape))
        else:
            shape=data.shape
            rValues=np.arange(0,shape[0])
            phiValues=np.zeros(shape[1]+1)
            phiValues[:-1]=np.arange(shape[1])*2*np.pi/(shape[1])
            
            r,phi=np.meshgrid(rValues,phiValues)

            if type(data)==Grid:
                data=data.array

                #        plt.pcolormesh(th, r, z)
        cmap=plt.get_cmap('inferno')
        if scale=='log':
            # data expected to come as r,phi but matplotlib wants phi,r
            #cmap.set_bad(cmap(0))
            cmap.set_bad(color='w')
            heatmap=ax.pcolormesh(phi,r,np.swapaxes(data,0,1),norm=matplotlib.colors.LogNorm(),cmap=cmap)
        else:
            heatmap=ax.pcolormesh(phi,r,np.swapaxes(data,0,1),cmap=cmap)
        fig.colorbar(heatmap,extend='both')
        apply_layout_to_ax(ax,layout)
        return fig
    
    @classmethod
    def show(cls,data,grid=False,layout={},scale='lin'):
        fig = cls.get_fig(data,grid=grid,layout=layout,scale=scale)
        fig.show()


        
class heat2D:
    @classmethod
    def get_fig(cls,data,grid=False,layout={},scale='lin',vmin = False,vmax =False,cmap='inferno'):   
        fig = plt.figure()
        ax = fig.add_subplot(111)
        #            cmap=plt.get_cmap('viridis')
        cmap=plt.get_cmap(cmap).copy()

        if not isinstance(grid,bool):
            grid=grid.copy()
            #        log.info('grid shape = {}'.format(grid.shape))
            x=grid[...,0]
            y=grid[...,1]
            #            log.info('phi shape={}'.format(phi.shape))
        else:
            shape=data.shape
            xValues=np.arange(0,shape[0])
            yValues=np.arange(0,shape[1])
            
            x,y=np.meshgrid(xValues,yValues)
            x=np.swapaxes(x,0,1)
            y=np.swapaxes(y,0,1)
            #log.debug('r shape = {} phi shape={}'.format(x.shape,y.shape))
            #        plt.pcolormesh(th, r, z)
        if scale=='log':
            cmap.set_bad(cmap(0))
            if isinstance(vmin,bool) and (not isinstance(vmax,bool)):
                vmin = data.min()
            elif isinstance(vmax,bool) and (not isinstance(vmin,bool)):
                vmax = data.min()
            elif isinstance(vmax,bool) and isinstance(vmin,bool):
                median_value = np.median(data)
                max_value = data.max()
                min_value = data.min()
                if (median_value>0) and (median_value!=np.nan) and (max_value<np.inf):
                    orders = int(np.log10(max_value/median_value)*0.8)
                    vmin,vmax = [median_value*10**(-orders//2),median_value*10**orders]
                elif (max_value==np.inf) and (median_value>0) and (median_value!=np.nan):            
                    orders = 12
                    vmin,vmax = [median_value*10**(-orders//2),median_value*10**orders]
                else:            
                    orders = 12
                    vmin,vmax = [max_value*10**(-orders//2),max_value]
            norm = matplotlib.colors.LogNorm(vmin=vmin,vmax=vmax)
            heatmap=ax.pcolormesh(x,y,data,norm = norm,cmap=cmap, shading ='auto')
        else:
            heatmap=ax.pcolormesh(x,y,data,cmap=cmap,vmin = vmin,vmax =vmax,shading='auto')
        fig.colorbar(heatmap)
        
        apply_layout_to_ax(ax,layout)
        return fig

    @classmethod
    def show(cls,data,grid=False,layout={},scale='lin',vmin = False,vmax =False):
        fig = cls.get_fig(data,grid=grid,layout=layout,scale=scale,vmin = vmin,vmax =vmax)
        fig.show()

class heat2D_multi(Plotter):
    @classmethod
    def get_fig(cls,datasets,grid=False,layout={},scale='lin',shape=(1,1),size = False,vmin = None,vmax =None,cmap = 'inferno'):
        cmap=plt.get_cmap(cmap).copy()
        if isinstance(size,bool):
            size = tuple(9*i for i in shape[::-1])
        #cmap=plt.get_cmap('viridis')        
        fig, axes = plt.subplots(nrows=shape[0], ncols=shape[1],figsize=size, gridspec_kw = {'wspace':0.2, 'hspace':0.45})
        if shape[0]==1:
            axes = [axes]
            datasets = [datasets]
        if shape[1]==1:
            axes = [[ax] for ax in axes]
            datasets = [[data] for data in datasets]

        if not isinstance(grid,bool):
            grid=grid.copy()
            #        log.info('grid shape = {}'.format(grid.shape))
            x=grid[...,0]
            y=grid[...,1]
            #            log.info('phi shape={}'.format(phi.shape))
        else:
            shape=flattenList(datasets)[0].shape
            xValues=np.arange(0,shape[0])
            yValues=np.arange(0,shape[1])
            
            x,y=np.meshgrid(xValues,yValues)
            x=np.swapaxes(x,0,1)
            y=np.swapaxes(y,0,1)
            #log.debug('r shape = {} phi shape={}'.format(x.shape,y.shape))
            #        plt.pcolormesh(th, r, z)
        heatmaps = []
        if scale=='log':
            cmap.set_bad(cmap(0))
            for ax_row,dataset_row in zip(axes,datasets):
                heatmap_row = []
                for ax,data in zip(ax_row,dataset_row):
                    #if isinstance(vmin,bool):
                    #    vmin = data.min()
                    #if isinstance(vmax,bool):
                    #    vmax = data.min()
                    heatmap_row.append(ax.pcolormesh(x,y,data,norm=matplotlib.colors.LogNorm(vmin=vmin,vmax=vmax),cmap=cmap,shading='auto'))
                    ax.set_aspect('equal')
                heatmaps.append(heatmap_row)
        else:
            for ax_row,dataset_row in zip(axes,datasets):
                heatmap_row = []
                for ax,data in zip(ax_row,dataset_row):
                    heatmap_row.append(ax.pcolormesh(x,y,data,cmap=cmap,vmin=vmin,vmax=vmax,shading='auto'))
                heatmaps.append(heatmap_row)
        #fig.colorbar(heatmap)
        for ax_row,heatmap_row in zip(axes,heatmaps):
            for ax,heatmap in zip(ax_row,heatmap_row):
                    fig.colorbar(heatmap , ax = ax)

        if isinstance(layout,(tuple,list)):
            for ax_row,dataset_row,layout_row in zip(axes,datasets,layout):
                for ax,data,l in zip(ax_row,dataset_row,layout_row):
                    apply_layout_to_ax(ax,l)
        else:
            for ax_row,dataset_row in zip(axes,datasets):
                for ax,data in zip(ax_row,dataset_row):
                    apply_layout_to_ax(ax,layout)
        return fig



class imshow:
    @classmethod
    def get_fig(cls,data,layout={},scale='lin',vmin = False,vmax =False,cmap = 'inferno',aspect = 1):     
        fig = plt.figure()
        ax = fig.add_subplot(111)
        cmap=plt.get_cmap(cmap).copy()
        if isinstance(vmin,bool):
            vmin = data.min()
        if isinstance(vmax,bool):
            vmax = data.min()
        if scale=='log':
            cmap.set_bad(cmap(0))
            heatmap=ax.imshow(data,norm=matplotlib.colors.LogNorm(vmin=vmin,vmax=vmax),cmap=cmap,aspect = aspect)
        else:
            heatmap=ax.imshow(data,cmap=cmap,vmin=vmin,vmax=vmax,aspect = aspect)
        fig.colorbar(heatmap)
        
        apply_layout_to_ax(ax,layout)
        return fig

    @classmethod
    def show(cls,data,layout={},scale='lin',vmin = False,vmax =False,cmap = 'inferno',aspect = 1):     
        fig = cls.get_fig(data,grid=grid,layout=layout,scale=scale,vmin = vmin,vmax =vmax,aspect = aspect,cmap=cmap)
        fig.show()


class agipd_heat_multi():
    @classmethod
    def get_fig(cls,densities,grid,print_mask,shape = (1,1),layouts=[{'x_label':r'$q_x \quad [\AA^-1] $','y_label':r'$q_y \quad [\AA^-1]$'}],scale='log',vmin=None,vmax=None):
        # assumes grid and density is ordered by modules
        cmap=plt.get_cmap('jet')
        fig, axes = plt.subplots(nrows=shape[0], ncols=shape[1],figsize=(9, 9), gridspec_kw = {'wspace':0.2, 'hspace':0.45})

        n_densities = len(densities)
        for ax in axes.ravel()[n_densities:]:
            ax.axis('off')
        ax_tuple = tuple(axes.flat)
        g_grid=grid[:,:-1,:-1][print_mask].reshape(-1,512,128,3)

        if scale=='log':
            norm = matplotlib.colors.LogNorm(vmin=vmin,vmax=vmax)
        else:
            morm = None
        base_shape = 100*shape[0]+10*shape[1]
        for id,density in enumerate(densities):
            ax = ax_tuple[id]
            print_data=np.zeros(print_mask.shape,dtype=float)
            print_data[print_mask]=density.flatten()
            for m_id in range(len(grid)):
                x = grid[m_id,:,:,0]
                y = grid[m_id,:,:,1]
                c = np.swapaxes(print_data[m_id],0,0)
                #heatmap = ax.pcolormesh(x, y, c, norm=norm ,cmap = cmap,vmin=density.min()+1e-10,vmax=density.max())
                heatmap = ax.pcolormesh(x, y, c, norm=norm ,cmap = cmap)
                ax.set_aspect(1)
            r = np.abs(x.max()-x.min())*0.01
            circ = plt.Circle((0,0), radius=r, linewidth=2, color='r')
            circ.set_fill(True)
            ax.add_patch(circ)
            
            apply_layout_to_ax(ax,layouts[id])
        #add color bar below chart
        #divider = make_axes_locatable(plt.gca())
        #cax = divider.append_axes("right", "5%", pad="3%")
        cax = fig.add_axes([0.85, 0.1, 0.075, 0.8])
        bar = fig.colorbar(ax.collections[0],ax = cax,fraction = 0.3 ,aspect = 30)
        cax.axis('off')
        fig.tight_layout()
        return fig

    @classmethod
    def show(cls,densities,grid,print_mask,shape = (1,1),layouts=[{'x_label':r'$q_x \quad [\AA^-1] $','y_label':r'$q_y \quad [\AA^-1]$'}],scale='log',vmin=None,vmax=None):
        fig = cls.get_fig(densities,grid,print_mask,shape = shape, layouts = layouts, scale = scale,vmin=vmin,vmax=vmax)
        fig.show()
                
class agipd_heat():
    @classmethod
    def get_fig(cls,density, grid,print_mask,fill_value = 0,gradients=False,layout={'x_label':r'$q_x \quad [\AA^-1] $','y_label':r'$q_y \quad [\AA^-1]$'},scale='log',vmin=None,vmax=None,cmap = 'viridis'):
        # assumes grid and density is ordered by modules
        cmap=plt.get_cmap(cmap).copy()
        fig = plt.figure()
        ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
               
        print_data=np.full(print_mask.shape,fill_value,dtype=float)#np.zeros(print_mask.shape)
        print_data[print_mask]=density.flatten()
        g_grid=grid[:,:-1,:-1][print_mask].reshape(-1,512,128,3)

        if scale=='log':
            norm = matplotlib.colors.LogNorm(vmin=vmin,vmax=vmax)
        else:
            norm = None
        log.info('gradients = {}'.format(gradients))
        
        for m_id in range(len(grid)):
            x = grid[m_id,:,:,0]
            y = grid[m_id,:,:,1]
            c = np.swapaxes(print_data[m_id],0,0)
            if (scale=='log') and (c.max()==0):
                continue
            #heatmap = ax.pcolormesh(x, y, c, norm=norm ,cmap = cmap,vmin=density[density>0].min()*0.1,vmax=density.max())
            heatmap = ax.pcolormesh(x, y, c, norm=norm ,cmap = cmap)
            if not isinstance(gradients,bool):
                log.info('fu8uuuu')
                X = g_grid[m_id,...,0]                
                Y = g_grid[m_id,...,1]
                U = gradients[m_id,...,0]
                V = gradients[m_id,...,1]
                q = ax.quiver(X, Y, U, V, scale=np.abs(gradients).max()*0.9, scale_units='inches')

            ax.set_aspect(1)
            r = np.abs(x.max()-x.min())*0.01
            circ = plt.Circle((0,0), radius=r, linewidth=2, color='r')
            circ.set_fill(True)
            ax.add_patch(circ)
        apply_layout_to_ax(ax,layout)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        bar = fig.colorbar(ax.collections[0],cax =cax)
        apply_layout_to_figure(fig,layout)
        return fig

    @classmethod
    def show(cls,density, grid,print_mask,fill_value = 0,gradients=False,layout={'x_label':r'$q_x \quad [\AA^-1]$','y_label':r'$q_y \quad [\AA^-1]$'},scale='log',vmin=None,vmax=None,cmap = 'inferno'):
        fig = cls.get_fig(density,grid,print_mask,fill_value=fill_value,gradients = gradients, layout = layout, scale = scale,vmin=vmin,vmax=vmax,cmap = cmap)
        fig.show()

class centering_heat:
    @classmethod
    def get_fig(cls,convolution,sigma_mask,center,errors,grid, layout_conv, layout_mask,scale='lin'):
     
        fig = plt.figure()
        ax_conv = fig.add_axes((0.05, 0.25, 0.4, 0.65))
        ax_mask = fig.add_axes((0.55, 0.25, 0.4, 0.65))
        #            cmap=plt.get_cmap('viridis')
        cmap=plt.get_cmap('inferno')

        if not isinstance(grid,bool):
            grid=grid.copy()
            #        log.info('grid shape = {}'.format(grid.shape))
            x=grid[...,0]
            y=grid[...,1]
            #            log.info('phi shape={}'.format(phi.shape))
        if scale=='log':
            cmap.set_bad(cmap(0))
            hm_conv=ax_conv.pcolormesh(x,y,convolution,norm=matplotlib.colors.LogNorm(),cmap=cmap)
            hm_mask=ax_mask.pcolormesh(x,y,sigma_mask,norm=matplotlib.colors.LogNorm(),cmap=cmap)
        else:
            hm_conv=ax_conv.pcolormesh(x,y,convolution,cmap=cmap)
            hm_mask=ax_mask.pcolormesh(x,y,sigma_mask,cmap=cmap)

        props = dict(boxstyle='round', facecolor='wheat', alpha=0.3)
        text = 'Center at: {}'.format(center)+'\n'+r'$e^{-\frac{1}{2}}$'+' errors: x {} Y {}'.format(errors[0],errors[1])      
        ax_mask.text(0.05, 0.05, text,horizontalalignment='left', verticalalignment='bottom',bbox=props, transform=ax_mask.transAxes)

            
        fig.suptitle(layout_conv["sup_title"],fontsize = 24)
        apply_layout_to_ax(ax_conv,layout_conv)
        apply_layout_to_ax(ax_mask,layout_mask)
        divider = make_axes_locatable(ax_conv)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        bar = fig.colorbar(ax_conv.collections[0],cax =cax)
        apply_layout_to_figure(fig,layout_conv)
        return fig

    @classmethod
    def show(cls,convolution,sigma_mask,center,errors,grid, layout_conv, layout_mask,scale='lin'):
        fig = cls.get_fig(convolution,sigma_mask,center,errors,grid, layout_conv, layout_mask,scale=scale)
        fig.show()
            

class plot1D:
    @classmethod
    def get_fig(cls,data_set,labels=False, grid = False,layout={},y_scale='lin',x_scale='lin',ylim=False,xlim=False,xticks=False,yticks=False):
        if isinstance(grid,bool):
            grid=np.arange(data_set.shape[-1])
        fig = plt.figure()
        ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
        if not isinstance(ylim,bool):
            ax.set_ylim(ylim)
        if not isinstance(xlim,bool):
            ax.set_xlim(xlim)
        if not isinstance(yticks,bool):
            ax.set_yticks(yticks)
        if not isinstance(xticks,bool):
            ax.set_xticks(xticks)
        if data_set.ndim>1:
            if not isinstance(labels,(list,tuple)):
                labels = [None]*len(data_set)
            colors = plt.get_cmap('viridis')(np.linspace(0,1,len(data_set)))
            for part,label,color in zip(data_set,labels,colors):
                #print(part.shape)
                #log.info('shapes grid = {} part = {} '.format(grid.shape,part.shape))
                #print('shapes grid = {} part = {} '.format(grid.shape,part.shape))
                ax.plot(grid,part,label = label,color = color)
        else:
            ax.plot(grid,data_set,label=labels[0])
        ax.grid()
        if isinstance(labels,(tuple,list,np.ndarray)):
            if len(labels)>=6:
                ax.legend(loc='lower center', bbox_to_anchor=(0.5, -0.2-0.05*len(labels)/6), ncol=6)
            else:
                ax.legend()
        else:
            ax.legend()
        if x_scale=='log':
            ax.set_xscale('log')
        if y_scale=='log':
            ax.set_yscale('log')
        apply_layout_to_ax(ax,layout)
        apply_layout_to_figure(fig,layout)
        return fig
    @classmethod
    def show(cls,data_set,labels=False,grid=False,layout={},x_scale='lin',y_scale='lin',ylim=False,xlim=False,xticks=False,yticks=False):
        fig = cls.get_fig(data_set,labels = labels, grid = grid, layout = layout, x_scale = x_scale, y_scale=y_scale,ylim=ylim,xlim=xlim,xticks=xticks,yticks=yticks)
        fig.show()

        
class scatter1D:
    @classmethod
    def get_fig(cls,data_set,s=None, c=None, marker=None,labels=False, grid = False,layout={},y_scale='lin',x_scale='lin',ylim=False,xlim=False,xticks=False,yticks=False):
        if isinstance(grid,bool):
            grid=np.arange(data_set.shape[-1])
        fig = plt.figure()
        ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
        if not isinstance(ylim,bool):
            ax.set_ylim(ylim)
        if not isinstance(xlim,bool):
            ax.set_xlim(xlim)
        if not isinstance(yticks,bool):
            ax.set_yticks(yticks)
        if not isinstance(xticks,bool):
            ax.set_xticks(xticks)
        if data_set.ndim>1:
            if not isinstance(labels,(list,tuple)):
                labels = [None]*len(data_set)
            colors = plt.get_cmap('viridis')(np.linspace(0,1,len(data_set)))
            for part,label,color in zip(data_set,labels,colors):
                print(part.shape)
                log.info('shapes grid = {} part = {} '.format(grid.shape,part.shape))
                #print('shapes grid = {} part = {} '.format(grid.shape,part.shape))
                ax.scatter(grid,part,s=s, c=c, marker=marker,label = label,color = color)
        else:
            ax.scatter(grid,data_set,s=s, c=c, marker=marker)
        ax.grid()
        #ax.legend(fontsize=26)
        if x_scale=='log':
            ax.set_xscale('log')
        if y_scale=='log':
            ax.set_yscale('log')
        apply_layout_to_ax(ax,layout)
        apply_layout_to_figure(fig,layout)
        return fig
    @classmethod
    def show(cls,data_set,labels=False,grid=False,layout={},x_scale='lin',y_scale='lin'):
        fig = cls.get_fig(data_set,labels = labels, grid = grid, layout = layout, x_scale = x_scale, y_scale=y_scale)
        fig.show()
                    
class bar1D:
    @classmethod
    def get_fig(cls,data_set,labels=False, grid = False,layout={},y_scale='lin',x_scale='lin',y_min = False):
        if isinstance(grid,bool):
            grid=np.arange(data_set.shape[0])
        log.info('grid type = {}'.format(type(grid)))
        fig = plt.figure()
        ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
        max_value = 0
        if data_set.ndim>1:
            if not isinstance(labels,(list,tuple)):
                labels = [None]*len(data_set)
            for part,label in zip(data_set,labels):
                ax.bar(grid,part,label = label,width = .95)
                max_value = np.max((max_value,np.max(part)))
        else:
            ax.plot(grid,data_set)
            max_value = np.max(data_set)
        if not isinstance(y_min,(bool)):
            ax.set_ylim(0,max_value)
        ax.legend(fontsize=16)
        if y_scale=='log':
            ax.set_yscale('log')
        if x_scale=='log':
            ax.set_xscale('log')

        apply_layout_to_ax(ax,layout)
        apply_layout_to_figure(fig,layout)
        return fig
    @classmethod
    def show(self,data_set,labels=False,grid=False,layout={},scale='lin'):
        fig = cls.get_fig(data_set,labels = labels, grid = grid, layout = layout, scale = scale)
        fig.show()

class hist1D:
    @classmethod
    def get_fig(cls,data_set,labels=False, grid = False,layout={},y_scale='lin',x_scale='lin',y_lim = False):
        log.info('y_lim={}'.format(y_lim))
        if isinstance(grid,bool):
            grid=np.arange(data_set.shape[0])
        log.info('grid type = {}'.format(type(grid)))
        fig = plt.figure()
        ax = fig.add_axes((0.1, 0.2, 0.8, 0.7))
        max_value = 0
        if data_set.ndim>1:
            if not isinstance(labels,(list,tuple)):
                labels = [None]*len(data_set)
            for part,label in zip(data_set,labels):
                ax.hist(part, bins='auto',label = make_string_tex_conform(label))
                max_value = np.max((max_value,np.max(part)))
        else:
            ax.hist(data_set,bins='auto')
            max_value = np.max(data_set)
        if isinstance(y_lim,(tuple,list)):
            ax.set_xlim(*y_lim) #on purpose since functionvalues are on x axis    
        ax.legend(fontsize=16)
        if y_scale=='log':
            ax.set_yscale('log')
        if x_scale=='log':
            ax.set_xscale('log')
        apply_layout_to_ax(ax,layout)
        apply_layout_to_figure(fig,layout)
        return fig
    @classmethod
    def show(cls,data_set,labels=False,grid=False,layout={},y_scale='lin'):
        fig = cls.get_fig(data_set,labels = labels, grid = grid, layout = layout, y_scale = y_scale)
        fig.show()        

class hist2D:
    @classmethod
    def get_fig(cls,data,bins = 10,labels=False, grid = False,layout={},y_scale='lin',x_scale='lin',range = None,norm = None,cmap ='viridis'):
        x=data[:,0]
        y=data[:,1]
        fig=plt.figure()
        ax=fig.add_subplot(111)
        cmap=plt.get_cmap(cmap).copy()
        if norm == 'log':
            cmap.set_bad(cmap(0))
            norm=matplotlib.colors.LogNorm()
            print('yay')
        hist = ax.hist2d(x,y,bins=bins,range = range,norm = norm,cmap=cmap)
        if y_scale=='log':
            ax.set_yscale('log')
        if x_scale=='log':
            ax.set_xscale('log')
        fig.colorbar(hist[3])
        apply_layout_to_ax(ax,layout)
        return fig
    @classmethod
    def show(cls,data_set,labels=False,grid=False,layout={},y_scale='lin'):
        fig = cls.get_fig(data_set,labels = labels, grid = grid, layout = layout, y_scale = y_scale)
        fig.show()        
        

class fxs_presenter:
    '''
    
    '''
    def applyLayoutToAx(self,ax,layout):
        title=layout.get('title',"Scatter plot 2D")
        title_fontsize=layout.get('title_fontsize',16)
        xLable=layout.get('x_label',"r")
        yLable=layout.get('y_label',"phi")
        label_fontsize=layout.get('label_fontsize',14)        
        ax.set_title(title,fontsize=title_fontsize)
        ax.set_xlabel(xLable,fontsize=label_fontsize)
        ax.set_ylabel(yLable,fontsize=label_fontsize)

    
    def present(self,plt_id,arg_dict):
#        log.info('arg dict={}'.format(arg_dict))
        if plt_id=='generic':            
            figures=self.generate_result_figures(arg_dict)
        elif plt_id=='reconstructed_densities':
            figures=self.generate_density_figures(arg_dict)
        elif plt_id=='reconstructed_densities_and_mask':
            figures=self.generate_density_mask_figures(arg_dict)
        elif plt_id=='reciprocal_projection_mask_sweep':
            figures=self.generate_reciprocal_projection_mask_sweep_figures(arg_dict)
        elif plt_id=='averaged_results':
            pass
        elif plt_id=='single_run_evolution':
            pass
        else:
            e=AssertionError('Unknown plot id: {} '.format(plt_id))
            log.error(e)
            raise
        return figures
        
    def present2(self):
        pass

    def generate_subplot(self,fig,ax_specifier,plot_dict):
        subplot_id=ax_specifier['key']
        if subplot_id=='real_density':
            self.draw_real_density_on_ax(fig,ax_specifier,plot_dict)
        elif subplot_id=='reciprocal_density':
            self.draw_reciprocal_density_on_ax(fig,ax_specifier,plot_dict)
        elif subplot_id=='support_mask':
            pass
        elif subplot_id=='error_metrics':
            self.draw_error_metric_on_ax(fig,ax_specifier)
        elif subplot_id=='harmonic_order_bCoeff':
            self.draw_bCoeff_on_ax(fig,ax_specifier,plot_dict)
            

    def generate_density_figures(self,data_dict):
        conf_dict=data_dict['conf_dict']
        harm_orders=conf_dict['fourier_transform']['harmonic_orders']
 
        
        plot_dict={}
        figures_data=[]
        n_datasets=len(data_dict['datasets'])
        title_part1=' (Run {} on {})'.format(data_dict.get('run','n/A'),data_dict.get('date','N/A'))
        title_part2=' consisting of {} HIO, {} ER and {} SW steps each. Using harmonic orders in $\{{ {},{},\ldots,{}\}}$'.format(conf_dict['main_loop']['nHIO'],conf_dict['main_loop']['nER'],conf_dict['main_loop']['nSW'],harm_orders.min(),harm_orders.min()+(harm_orders[1]-harm_orders[0]),harm_orders.max())
        for num,dataset in enumerate(data_dict['datasets']):
            axes_data={}
            axes_data['real_density']=[dataset['real_density'].real,{'data':dataset['initial_density'],'title':'Initial Density'}]
            axes_data['reciprocal_density']={'data':np.square(np.abs(dataset['reciprocal_density'])),'title':'Intensity'}
            axes_data['error_list']={'data':dataset['error_list'],'title':'Error Metric'}
            figure_data={'axes_data':axes_data}
            figure_data['title']='Reconstruction Results {} of {}'.format(num,n_datasets,plot_dict.get('run','n/A'),plot_dict.get('date','N/A'))+title_part1+'\n {} loops'.format(dataset['loop_iterations'])+title_part2
            figures_data.append(figure_data)
        plot_dict['figures_data']=figures_data
        plot_dict['default_figure_shape']=(int(2),int(2))
        plot_dict['default_subplot_positions']={'real_density':['221','223'],'reciprocal_density':['222'],'error_list':['224']}
        plot_dict['grid_pair']=conf_dict['internal_grids']
        figures=self.generate_figures(plot_dict)
        return figures

    def generate_density_mask_figures(self,data_dict):
        conf_dict=data_dict['conf_dict']
        harm_orders=conf_dict['fourier_transform']['harmonic_orders']
 #       log.info('configuration dict keys=\n {}'.format(conf_dict.keys()))
        
        plot_dict={}
        figures_data=[]
        n_datasets=len(data_dict['datasets'])
        title_part1=' (Run {} on {})'.format(data_dict.get('run','n/A'),data_dict.get('date','N/A'))
        title_part2=' consisting of {} HIO, {} ER and {} SW steps each. Using harmonic orders in $\{{ {},{},\ldots,{}\}}$'.format(conf_dict['main_loop']['HIO']['loop_iterations'],conf_dict['main_loop']['ER']['loop_iterations'],conf_dict['main_loop']['SW']['loop_iterations'],harm_orders.min(),harm_orders.min()+(harm_orders[1]-harm_orders[0]),harm_orders.max())
        for num,dataset in enumerate(data_dict['datasets']):
#            log.info('plot dataset ={}'.format(dataset))
#            log.info('dataset type ={}'.format(type(dataset)))
#            log.info('dataset shape={}'.format(dataset.shape))
#            log.info('dataset[0] type={}'.format(type(dataset[0])))
            
            axes_data={}
            axes_data['real_density']=[dataset['real_density'].real,{'data':dataset['support_mask'].astype(int),'title':'Support Mask'}]
            axes_data['reciprocal_density']={'data':np.square(np.abs(dataset['reciprocal_density'])),'title':'Intensity'}
            axes_data['error_list']={'data':dataset['error_list'],'title':'Error Metric'}
            figure_data={'axes_data':axes_data}
            figure_data['title']='Reconstruction Results {} of {}'.format(num,n_datasets,plot_dict.get('run','n/A'),plot_dict.get('date','N/A'))+title_part1+'\n {} loops'.format(dataset['loop_iterations'])+title_part2
            figures_data.append(figure_data)
        plot_dict['figures_data']=figures_data
        plot_dict['default_figure_shape']=(int(2),int(2))
        plot_dict['dec@fault_subplot_positions']={'real_density':['221','223'],'reciprocal_density':['222'],'error_list':['224']}
        plot_dict['grid_pair']=conf_dict['internal_grids']
        figures=self.generate_figures(plot_dict)
        return figures
    def generate_error_metric_figure(self,data_dict):
        grids=data_dict['configuration']['internal_grid']
        opt=data_dict['configuration']['opt']
        #       log.info('configuration dict keys=\n {}'.format(conf_dict.keys()))
        plt.rcParams.update({'font.size': 26})
        plot_dict={}
        figures_data=[]
        n_datasets=len(data_dict['reconstruction_results'])
        name = str(opt['name'])[2:-1].replace('_','\_')
        title_part1='{}\n {},  Run {}  \n'.format(name,data_dict.get('date','N/A').replace('_','\_'),data_dict.get('run','n/A'))
        log.info('reconstruction_result keys = {}'.format(data_dict['reconstruction_results'].keys()))
        for num,dataset in data_dict['reconstruction_results'].items():
#            log.info('plot dataset ={}'.format(dataset))
#            log.info('dataset type ={}'.format(type(dataset)))
#            log.info('dataset shape={}'.format(dataset.shape))
#            log.info('dataset[0] type={}'.format(type(dataset[0])))
            
            axes_data={}            
            axes_data['error_metrics']={'data':dataset['error_dict'],'title':'Error Metrics'}
            figure_data={'axes_data':axes_data}
            figure_data['title']=title_part1+'Reconstruction Results {} of {}'.format(int(num)+1,n_datasets)
            figures_data.append(figure_data)
        plot_dict['figures_data']=figures_data
        plot_dict['default_figure_shape']=(int(1),int(1))
        plot_dict['dec@fault_subplot_positions']={'error_metrics':['111']}
        plot_dict['grid_pair']=grids
        figures=self.generate_figures(plot_dict)
        return figures

    def generate_reciprocal_projection_mask_sweep_figures(self,data_dict):
        conf_dict=data_dict['conf_dict']
        masks=conf_dict['multi_process']['arg_dict']['reciprocal_projection']['mask']
        harm_orders=conf_dict['used_harmonic_orders']
        log.info('configuration dict keys=\n {}'.format(conf_dict.keys()))
        
        plot_dict={}
        figures_data=[]
        n_datasets=len(data_dict['datasets'])
        title_part1=' (Run {} on {})'.format(data_dict.get('run','n/A'),data_dict.get('date','N/A'))
        title_part2=' consisting of {} HIO, {} ER and {} SW steps each + {} ER refinment steps. Using harmonic orders in $\{{ {},{},\ldots,{}\}}$'.format(conf_dict['main_loop']['HIO_iterations'],conf_dict['main_loop']['ER_iterations'],conf_dict['main_loop']['SW_iterations'],conf_dict['main_loop']['ER_refinement_iterations'],harm_orders.min(),harm_orders.min()+(harm_orders[1]-harm_orders[0]),harm_orders.max())
        for num,dataset in enumerate(data_dict['datasets']):
            axes_data={}
            axes_data['real_density']=[dataset['real_density'].real,{'data':dataset['support_mask'].astype(int),'title':'Support Mask'}]
            reciprocal_density=np.square(np.abs(dataset['reciprocal_density']))
            reciprocal_density[~masks[num]]*=1e-3
            axes_data['reciprocal_density']={'data':reciprocal_density,'title':'Intensity'}
            axes_data['error_list']={'data':dataset['error_list'],'title':'Error Metric'}
            figure_data={'axes_data':axes_data}
            figure_data['title']='Reconstruction Results {} of {}'.format(num,n_datasets,plot_dict.get('run','n/A'),plot_dict.get('date','N/A'))+title_part1+'\n {} loops'.format(dataset['loop_iterations'])+title_part2
            figures_data.append(figure_data)
        plot_dict['figures_data']=figures_data
        plot_dict['figure_shape']=(int(2),int(2))
        plot_dict['default_subplot_positions']={'real_density':['221','223'],'reciprocal_density':['222'],'error_list':['224']}
        plot_dict['grid_pair']=conf_dict['internal_grids']
        figures=self.generate_figures(plot_dict)
        return figures
    

    def generate_result_figures(self,plot_dict):
        figures=self.generate_figures(plot_dict)
        return figures


    def generate_figures(self,plot_dict):
        def calc_plot_shape(n_subplots):
            if n_subplots==1:
                shape=(1,1)
            elif n_subplots==2:
                shape=(1,2)
            else:
                n_rows_prelim=np.floor(np.sqrt(n_subplots))
                n_rows=n_rows_prelim+n_rows_prelim%2
                n_cols=np.ceil(n_subplots/n_rows)
                shape=(int(n_rows),int(n_cols))
#            log.info('shape={}'.format(shape))
            return shape

        def calc_n_subplots_per_key(axes_data,keys):
            n_subplots_per_key={}
            for key in keys:
                dataset=axes_data[key]
                if isinstance(dataset,(list,tuple)):
                    n_subplots_per_key[key]=len(dataset)
                else:
                    n_subplots_per_key[key]=1
#            log.info('n_subplots_per_key={}'.format(n_subplots_per_key))
            return n_subplots_per_key
        
        def generate_subplot_positions(plot_shape,n_subplots_per_key):
            n_rows=plot_shape[0]
            n_assigned_positions=0
            subplot_positions={}            
            for key in n_subplots_per_key.keys():
                subplot_positions[key]=[]                
                for subplot_num in range(1,n_subplots_per_key[key]+1):
                    num=str(n_assigned_positions+subplot_num)
                    subplot_positions[key].append(str(plot_shape[0])+str(plot_shape[1])+num)
                n_assigned_positions+=n_subplots_per_key[key]
#            log.info('subplot_positions=\n{}'.format(subplot_positions))   
            return subplot_positions
        def extract_figures_data(plot_dict):        
            figure_data=plot_dict.get('figure_data',False)
            if not isinstance(figure_data,bool):
                figures_data=[figure_data]
            else:
                figures_data=plot_dict['figure_data']
            return figures_data

        def select_first_non_bool(*args):
            selection=False
            for arg in args:
                if not isinstance(arg,bool):
                    selection=arg
                    break
            return selection

        def call_generate_subplot(key,ax_position,dataset,ax_titles,fig,plot_dict):
            if isinstance(dataset,dict):
                data=dataset.pop('data')
                ax_options=dataset                
            else:
                data=dataset
                ax_options={}
                
            if not isinstance(ax_titles,bool):
                title_by_pos=ax_titles.get(ax_position,False)
#                log.info('axoptions = {}'.format(ax_options))
                ax_options['title_by_pos']=title_by_pos
#            log.info('ax_option keys={}'.format(ax_options.keys()))
            ax_specifier={'key':key,'dataset':data,'position':ax_position,'options':ax_options}
            self.generate_subplot(fig,ax_specifier,plot_dict)
            
        figures_data=extract_figures_data(plot_dict)
        default_figure_shape=plot_dict.get('default_figure_shape',False)
        default_datakeys_to_plot=plot_dict.get('default_datakeys_to_plot',False)
        default_subplot_positions=plot_dict.get('default_subplot_positions',False)
        default_title=plot_dict.get('default_title',False)
        
#        plt.rcParams['font.size'] = plot_dict.get('default_font_size',16)
        figure_list=[]
        for figure_data in figures_data:
            axes_data=figure_data['axes_data']
            datakeys_to_plot=select_first_non_bool(figure_data.get('datakeys_to_plot',False),default_datakeys_to_plot,list(axes_data))
            n_subplots_per_key=calc_n_subplots_per_key(axes_data,datakeys_to_plot)
            n_subplots=sum(n_subplots_per_key.values())
            shape=select_first_non_bool(figure_data.get('shape',False),default_figure_shape,calc_plot_shape(n_subplots))
            try:
                assert np.prod(shape)>=n_subplots,'Specified plot shape {} conflicts with number of subplots:{}. Using autogenerated shape!'.format(shape,n_subplots)
            except AssertionError as e:
                log.error(e)
                shape=calc_plot_shape(n_subplots)            
            subplot_positions=select_first_non_bool(figure_data.get('subplot_positions',False),default_subplot_positions,generate_subplot_positions(shape,n_subplots_per_key))
            title=select_first_non_bool(figure_data.get('title',False),default_title,'')
            ax_titles=select_first_non_bool(plot_dict.get('default_ax_titles',False),figure_data.get('ax_titles',False))
            
            fig=plt.figure(figsize=(20,20))
            fig.suptitle(title)
            for key,datasets in figure_data['axes_data'].items():
#                log.info('key={}'.format(key))
                subplot_positions_per_key=subplot_positions[key].copy()
                if isinstance(datasets,(list,tuple)):
                    for dataset in datasets:
                        ax_position=subplot_positions_per_key.pop(0)
                        call_generate_subplot(key,ax_position,dataset,ax_titles,fig,plot_dict)
                else:
                    dataset=datasets
                    ax_position=subplot_positions_per_key.pop(0)
                    call_generate_subplot(key,ax_position,dataset,ax_titles,fig,plot_dict)
            figure_list.append(fig)
        return figure_list

    
    def draw_1D_funciton_on_ax(self,data,fig,ax,layout={},scale='lin'):
#        log.info('1D data=\n{}'.format(data))
        ax.plot(np.arange(data.shape[0]),data)
        self.applyLayoutToAx(ax,layout)
        if scale=='log':
            ax.set_yscale('log')
            
    def draw_polar_heatplot_on_ax(self,data,fig,ax,layout={},scale='lin',grid=False):
        #            cmap=plt.get_cmap('viridis')
        cmap=plt.get_cmap('inferno')

        if not isinstance(grid,bool):
            grid=grid.copy()
            #        log.info('grid shape = {}'.format(grid.shape))
            grid.array=np.concatenate((grid[:],np.expand_dims(grid[:,0],1)),axis=1)
            r=np.swapaxes(grid[...,0],0,1)
            phi=np.swapaxes(grid[...,1],0,1)
            #            log.info('phi shape={}'.format(phi.shape))
        else:
            shape=data.shape
            rValues=np.arange(0,shape[0])
            phiValues=np.zeros(shape[1]+1)
            phiValues[:-1]=np.arange(0,shape[1])*2*np.pi/(shape[1])            
            r,phi=np.meshgrid(rValues,phiValues)
            
        if type(data)==Grid:
            data=data.array
                
        if scale=='log':
            cmap.set_bad(cmap(0))
            heatmap=ax.pcolormesh(phi,r,np.swapaxes(data,0,1),norm=matplotlib.colors.LogNorm(),cmap=cmap)
        else:
            heatmap=ax.pcolormesh(phi,r,np.swapaxes(data,0,1),cmap=cmap)
#        log.info(' last r value = {}'.format(r[-1]))
        ax.set_yticks([r[-1,-1]])
        axins = inset_axes(ax,
                   width="5%",  # width = 5% of parent_bbox width
                   height="90%",  # height : 50%
                   loc='center left',
                   bbox_to_anchor=(1.05, 0., 1, 1),
                   bbox_transform=ax.transAxes,
                   borderpad=0,
                   )
        fig.colorbar(heatmap,cax=axins,ax=ax)        
        self.applyLayoutToAx(ax,layout) 

    def draw_heatplot_on_ax(self,data,fig,ax,layout={},scale='lin',grid=False):
        #            cmap=plt.get_cmap('viridis')
        cmap=plt.get_cmap('inferno')
        if not isinstance(grid,bool):
            grid=Grid.copy(grid)
            #        log.info('grid shape = {}'.format(grid.shape))
            x=np.swapaxes(grid[:][:,0].reshape(grid.shape),0,1)
            y=np.swapaxes(grid[:][:,1].reshape(grid.shape),0,1)
            #            log.info('phi shape={}'.format(phi.shape))
        else:
            shape=data.shape
            xValues=np.arange(0,shape[0])
            yValues=np.arange(0,shape[1])            
            x,y=np.meshgrid(xValues,yValues)
            
        if type(data)==Grid:
            data=data.array
#        log.info('bcoeff data shape=\n{}'.format(data.shape))
        if scale=='log':
            cmap.set_bad(cmap(0))
            heatmap=ax.pcolormesh(x,y,np.swapaxes(data,0,1),norm=matplotlib.colors.LogNorm(),cmap=cmap)
        else:
            heatmap=ax.pcolormesh(x,y,np.swapaxes(data,0,1),cmap=cmap)
        fig.colorbar(heatmap,ax=ax)        
        self.applyLayoutToAx(ax,layout) 
        
    def draw_error_metric_on_ax(self,fig,ax_specifier):
        position=ax_specifier['position']
        title=ax_specifier['options'].get('title',False)
        std_error=ax_specifier['dataset']['real_rel_mean_square']
        cauchy_real=ax_specifier['dataset']['cauchy_real']
        cauchy_reciprocal=ax_specifier['dataset']['cauchy_reciprocal']
        _max=cauchy_real.max()
        cauchy_real/=_max
        cauchy_reciprocal/=_max
        cauchy_real[:2]=[2,1]
        cauchy_reciprocal[:2]=[2,1]

        x=np.arange(len(std_error))
        ax=fig.add_subplot()
        layout={'title':'','x_label':'Number of Reconstruciton Steps','y_label':'Error Estimate','title_fontsize':26,'label_fontsize':26}
        if not isinstance(title,bool):
            layout['title']=title
        
        ax.plot(x,cauchy_real,'#20F0A0')
        ax.plot(x,cauchy_reciprocal,'#8080F0')
        ax.plot(x,std_error,'k', linewidth=3)
        ax.legend(('Real: Cauchy', 'Reciprocal: Cauchy','Real: Relative Squared Intensity'),
           loc='upper right')
        self.applyLayoutToAx(ax,layout)
        ax.set_yscale('log')


        
    def draw_real_density_on_ax(self,fig,ax_specifier,plot_dict):
        position=ax_specifier['position']
        title=ax_specifier['options'].get('title',False)
        dataset=ax_specifier['dataset']
        log.info('dataset shape = {}'.format(dataset.shape))
        
        ax=fig.add_subplot(position,projection='polar')
        ax.set_yticklabels([])
        ax.set_xticklabels([])
        grid_pair=plot_dict.get('grid_pair',False)
        if isinstance(grid_pair,bool):
            realGrid=False
        else:
            realGrid=grid_pair.realGrid
        layout={'title':'Real Density','x_label':'','y_label':'','title_fontsize':12,'label_fontsize':10}
        if not isinstance(title,bool):
            layout['title']=title
        #log.info('realGrid.shape={},dataset shape = {}'.format(realGrid.shape,dataset.array.shape))
        self.draw_polar_heatplot_on_ax(dataset,fig,ax,layout=layout,grid=realGrid)

    def draw_reciprocal_density_on_ax(self,fig,ax_specifier,plot_dict):
        position=ax_specifier['position']
        title=ax_specifier['options'].get('title',False)
        dataset=ax_specifier['dataset']
        dataset=np.where(dataset<=0,1e-12,dataset)
        
        ax=fig.add_subplot(position,projection='polar')
        ax.set_yticklabels([])
        ax.set_xticklabels([])
        grid_pair=plot_dict.get('grid_pair',False)
        if isinstance(grid_pair,bool):
            reciprocalGrid=False
        else:
            reciprocalGrid=grid_pair.reciprocalGrid
            
        layout={'title':'reciprocal density','x_label':'','y_label':'','title_fontsize':12,'label_fontsize':10}
        if not isinstance(title,bool):
            layout['title']=title
        #log.info('realGrid.shape={},dataset shape = {}'.format(realGrid.shape,dataset.array.shape))
        self.draw_polar_heatplot_on_ax(dataset,fig,ax,layout=layout,grid=reciprocalGrid,scale='log')

    def draw_bCoeff_on_ax(self,fig,ax_specifier,plot_dict):
        options=ax_specifier['options']
        order=options['order']
        position=ax_specifier['position']
        title=options.get('title',False)        
        dataset=ax_specifier['dataset']
        
        ax=fig.add_subplot(position)
        grid_pair=plot_dict.get('grid_pair',False)
        if isinstance(grid_pair,bool):
            reciprocalGrid=False
        else:
            reciprocalGrid=grid_pair.reciprocalGrid
            coords=reciprocalGrid[:,0][:,0]
            grid=getGridByXYValues(coords,coords)
            
        
            
        layout={'title':'B-Coefficients','x_label':'$q_1$  []','y_label':'$q_2$  []','title_fontsize':12,'label_fontsize':10}
        if not isinstance(title,bool):
            layout['title']=title
#        log.info('data shape = {}'.format(dataset.shape))
#        log.info('grid shape = {}'.format(grid.array.shape))
        self.draw_heatplot_on_ax(dataset,fig,ax,layout=layout,scale='log')
        
    def generate_reciprocal_density_plot(self,dataset,fig,ax):        
        pass



import abc
import logging
import traceback
from xframe.startup_routines import _inject_dependency


log=logging.getLogger('root')

class ProjectWorkerInterface(abc.ABC):
    def __init__(self):
        from xframe.settings import project as settings
        from xframe.database import project as db
        from xframe.Multiprocessing import comm_module as comm
        self.settings = settings
        self.db = db
        self.comm = comm
        
    @abc.abstractclassmethod
    def run(self,*argr):
        pass

class ExperimentWorkerInterface(abc.ABC):
    """
    Abstract class for Experiments.    
    """
    def __init__(self):
        from xframe.settings import experiment as settings
        from xframe.database import experiment as db
        from xframe.Multiprocessing import comm_module as comm
        self.settings = settings
        self.db = db
        self.comm = comm
    
    @abc.abstractmethod
    def get_data(self,mode:str,range:dict,data_modifier : object = 'None'):
        pass
    @abc.abstractmethod
    def get_pixel_grid_reciprocal(self,coord_system:str):
        pass
    @abc.abstractmethod
    def get_geometry(self):
        pass
    @abc.abstractmethod
    def run(self):
        pass           
    
class CommunicationInterface(abc.ABC):
    @abc.abstractmethod
    def get_data(self):        
        pass
    @abc.abstractmethod
    def get_geometry(self):        
        pass
    @abc.abstractmethod
    def request_mp_evaluation(self,func,**kwargs):
        pass

class DatabaseInterface(abc.ABC):
    @abc.abstractmethod
    def load(self,analysisType,restultId):
        pass
    @abc.abstractmethod
    def save(self,analysisType,restultId):
        pass

class PresenterInterface(abc.ABC):
    @abc.abstractmethod
    def present(self):
        pass
    @abc.abstractmethod
    def save(self,path):
        pass


def error_message(cls):
    return f'Optional Dependency {cls._external_dependency_name_} missing.'

class DependencyImportError(Exception):
    def __init__(self,message=""):
        super().__init__(message)
        
class DependencyMeta(abc.ABCMeta):
    '''
    If this meta class is used by a class it will behave as follows.
    After Class creation the __call__, __getattr__ and __getattribute__ are overwritten.
    This cause any further interaction e.g calling the class or getting access to attributes
    trigger an attempt to load the actual external dependency.
    '''
    def __init__(cls,*args,**kwargs):
        super().__init__(*args,**kwargs)
        #new_meta._arm(cls)
        cls._armed = True
        
    def _try_to_import_dependency(cls,*args,**kwargs):
        dependency_name = cls._external_dependency_name_
        try:
            dependency = _inject_dependency(dependency_name)
        except Exception as e:            
            message = f'Requested external dependency "{dependency_name}" could not be imported.'
            log.error(message)
            raise DependencyImportError(message) from e
        return dependency
    

    def __call__armed(cls,*args,**kwargs):
        instance = cls._try_to_import_dependency()(*args,**kwargs)
        return instance
    def __getattr__armed(cls,key):
        if key[0]!='_':
            dep = cls._try_to_import_dependency()
            attr = getattr(dep,key)
            return attr
    def __getattribute__armed(cls,key):
        if key[0]!='_':
            dep = cls._try_to_import_dependency()
            attr = getattr(dep,key)
        else:
            attr = super().__getattribute__(key)
        return attr

    
    def __call__(cls,*args,**kwargs):
        if not cls._armed:
            return super().__call__(*args,**kwargs)
        else:
            return cls.__call__armed(*args,**kwargs)
    def __getattr__(cls,key):
         if key[0]!='_':
            if cls._armed:
                return cls.__getattr__armed(key)
            
    def __getattribute__(cls,key):
        attr = super().__getattribute__(key)
        if key[0]!='_':          
            if cls._armed:
              return cls.__getattr__armed(key)
        return attr


class ClickInterface(abc.ABC):
    _external_dependency_name_ = 'click'
    @property
    @abc.abstractmethod
    def click(self):
        pass    
class ClickDependency(ClickInterface,metaclass=DependencyMeta):
    pass

class DetectorInterface(abc.ABC):
    @abc.abstractmethod
    def get_geometry(self):
        pass
